from __future__ import annotations

import logging
import re

from soda.cloud.dbt_config import DBT_CLOUD_FALLBACK_ACCESS_URL, DbtCloudConfig
from soda.cloud.soda_cloud import SodaCloud
from soda.common.logs import Logs
from soda.common.parser import Parser
from soda.configuration.configuration import Configuration
from soda.sampler.default_sampler import DefaultSampler
from soda.sampler.http_sampler import HTTPSampler
from soda.sampler.soda_cloud_sampler import SodaCloudSampler

logger = logging.getLogger(__name__)

DATA_SOURCE = "data_source"
CONNECTION = "connection"
SODA_CLOUD = "soda_cloud"
DBT_CLOUD = "dbt_cloud"


class ConfigurationParser(Parser):
    def __init__(self, configuration: Configuration, logs: Logs, file_path: str):
        super().__init__(file_path=file_path, logs=logs)

        self.configuration: Configuration = configuration

    def parse_environment_yaml_str(self, environment_yaml_str: str):
        environment_yaml_str = self._resolve_jinja(environment_yaml_str)
        environment_dict = self._parse_yaml_str(environment_yaml_str)
        if environment_dict is not None:
            self.__parse_headers(environment_dict)

    def __parse_headers(self, environment_dict: dict) -> None:
        if not environment_dict:
            return

        for environment_header, header_value in environment_dict.items():
            if environment_header.startswith(f"{DATA_SOURCE} "):
                self._push_path_element(environment_header, header_value)
                data_source_name = environment_header[len(f"{DATA_SOURCE} ") :].strip()
                if not re.compile(r"^[a-z_][a-z_0-9]+$").match(data_source_name):
                    self.logs.error(
                        f"Invalid data source name '{data_source_name}'. Data source names must "
                        f"start with a lower case char or an underscore [a-z_], followed by any "
                        f"number of lower case chars, digits or underscore [a-z0-9_]"
                    )
                self.configuration.data_source_properties_by_name[data_source_name] = header_value

                self._get_required("type", str)

                # Backward compatibility. Merge connection properties one level up, into the data source properties. DS properties take precedence.
                data_source_connection = header_value.get("connection")
                if data_source_connection:
                    if isinstance(data_source_connection, dict):
                        for k, v in data_source_connection.items():
                            if k not in self.configuration.data_source_properties_by_name[data_source_name]:
                                self.configuration.data_source_properties_by_name[data_source_name][k] = v
                    else:
                        self.logs.error(
                            "'connection' configuration must be a dict",
                            location=self.location,
                        )

                # CORE-455 only parse sampler config if the data source name in file matches the scan data source name
                if data_source_name == self.configuration.scan._data_source_name:
                    self.parse_sampler_config(header_value)

                self._pop_path_element()

            elif environment_header == SODA_CLOUD:
                self._push_path_element(SODA_CLOUD, header_value)
                self.configuration.soda_cloud = self.parse_soda_cloud_cfg(header_value)
                self._pop_path_element()

            elif environment_header == DBT_CLOUD:
                self._push_path_element(DBT_CLOUD, header_value)
                self.configuration.dbt_cloud = self.parse_dbt_cloud_cfg(header_value)
                self._pop_path_element()

            else:
                self.logs.error(
                    f'Invalid configuration header: expected "{DATA_SOURCE} {{data source name}}".',
                    location=self.location,
                )

    def parse_sampler_config(self, header_value):
        # Sampler configuration
        disable_samples = None
        if "disable_samples" in header_value.keys():
            disable_samples = header_value.get("disable_samples")
            self.logs.info(
                "Data source configuration key `disable_samples` will be removed in future versions of Soda Core. Use data_source.sampler.disable_samples instead."
            )
        sampler_configuration = header_value.get("sampler")
        if sampler_configuration:
            if isinstance(sampler_configuration, dict):
                disable_samples = sampler_configuration.get("disable_samples", disable_samples)
                exclude_columns = sampler_configuration.get("exclude_columns", {})
                if exclude_columns:
                    if isinstance(exclude_columns, dict):
                        self.configuration.exclude_columns = exclude_columns
                    else:
                        self.logs.error(
                            "'exclude_columns' configuration must be a dict",
                            location=self.location,
                        )
                samples_limit = sampler_configuration.get("samples_limit")
                if samples_limit:
                    if isinstance(samples_limit, int):
                        self.configuration.samples_limit = samples_limit
                    else:
                        self.logs.error(
                            "'samples_limit' configuration must be an int",
                            location=self.location,
                        )

                storage = sampler_configuration.get("storage")
                if storage:
                    storage_type = storage.get("type")
                    if storage_type in ["http", "s3"]:
                        if storage_type == "http":
                            url = storage.get("url")
                            message = storage.get("message") or f"Failed rows have been sent to {url}"
                            link_text = storage.get("link_text") or message
                            link = storage.get("link")
                            self.configuration.sampler = HTTPSampler(
                                url,
                                message=message,
                                link=link,
                                link_text=link_text,
                            )
                    elif self.configuration.soda_cloud and not disable_samples:
                        self.configuration.sampler = SodaCloudSampler()
                    else:
                        self.logs.error(
                            f"Invalid storage type: {storage_type} specified, must be one of ['http', 's3'], using Soda Cloud as sampler",
                            location=self.location,
                        )

            else:
                self.logs.error(
                    "'sampler' configuration must be a dict",
                    location=self.location,
                )
        if disable_samples:
            self.configuration.sampler = DefaultSampler()

    def parse_soda_cloud_cfg(self, config_dict: dict):
        api_key = config_dict.get("api_key_id")
        api_secret = config_dict.get("api_key_secret")
        host = None
        if "host" in config_dict:
            host = config_dict.get("host")
        port = None
        if "port" in config_dict:
            port = config_dict.get("port")
        scheme = None
        if "scheme" in config_dict:
            scheme = config_dict.get("scheme")
        return SodaCloud(
            api_key_id=api_key,
            api_key_secret=api_secret,
            host=host,
            token=None,
            port=port,
            logs=self.logs,
            scheme=scheme,
        )

    def parse_dbt_cloud_cfg(self, config_dict: dict):
        api_token = config_dict.get("api_token")
        account_id = config_dict.get("account_id")
        access_url = config_dict.get("access_url", DBT_CLOUD_FALLBACK_ACCESS_URL)
        api_url = config_dict.get("api_url")

        if api_url:
            raise AttributeError(
                "The 'api_url' property is now deprecated. "
                f"If you wish to provide a different base URL than {DBT_CLOUD_FALLBACK_ACCESS_URL} "
                "use the 'access_url' property instead. See https://go.soda.io/access-url for more information"
            )

        return DbtCloudConfig(api_token=api_token, account_id=account_id, access_url=access_url)



from __future__ import annotations

from soda.sodacl.location import Location
from soda.sodacl.metric_check_cfg import MetricCheckCfg
from soda.sodacl.missing_and_valid_cfg import MissingAndValidCfg
from soda.sodacl.threshold_cfg import ThresholdCfg


class ChangeOverTimeMetricCheckCfg(MetricCheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: str | None,
        location: Location,
        name: str | None,
        metric_name: str,
        metric_args: list[object] | None,
        missing_and_valid_cfg: MissingAndValidCfg | None,
        filter: str | None,
        condition: str | None,
        metric_expression: str | None,
        metric_query: str | None,
        change_over_time_cfg: ChangeOverTimeCfg | None,
        fail_threshold_cfg: ThresholdCfg | None,
        warn_threshold_cfg: ThresholdCfg | None,
        samples_limit: int | None = None,
    ):
        super().__init__(
            source_header,
            source_line,
            source_configurations,
            location,
            name,
            metric_name,
            metric_args,
            missing_and_valid_cfg,
            filter,
            condition,
            metric_expression,
            metric_query,
            change_over_time_cfg,
            fail_threshold_cfg,
            warn_threshold_cfg,
            samples_limit=samples_limit,
        )



from __future__ import annotations

from soda.execution.query.query import Query
from soda.execution.query.sample_query import SampleQuery


class ReferenceQuery(Query):
    @staticmethod
    def build_source_column_list(metric):
        source_column_names = metric.check.check_cfg.source_column_names
        return ",".join(source_column_names)

    def __init__(
        self,
        data_source_scan: DataSourceScan,
        metric: ReferentialIntegrityMetric,
        partition: Partition,
        samples_limit: int | None = None,
    ):
        super().__init__(
            data_source_scan=data_source_scan,
            unqualified_query_name=f"reference[{ReferenceQuery.build_source_column_list(metric)}]",
            samples_limit=samples_limit,
            partition=partition,
        )

        self.metric = metric

        from soda.execution.data_source import DataSource
        from soda.sodacl.reference_check_cfg import ReferenceCheckCfg

        data_source: DataSource = data_source_scan.data_source

        check_cfg: ReferenceCheckCfg = metric.check.check_cfg
        source_table_name = data_source.qualified_table_name(metric.partition.table.table_name)
        source_column_names = check_cfg.source_column_names
        target_table_name = data_source.qualified_table_name(check_cfg.target_table_name)
        target_column_names = check_cfg.target_column_names

        selectable_source_columns = self.data_source_scan.data_source.sql_select_all_column_names(
            self.partition.table.table_name
        )
        source_diagnostic_column_fields = ", ".join([f"SOURCE.{c}" for c in selectable_source_columns])

        # TODO add global config of table diagnostic columns and apply that here
        # source_diagnostic_column_names = check_cfg.source_diagnostic_column_names
        # if source_diagnostic_column_names:
        #     source_diagnostic_column_names += source_column_names
        #     source_diagnostic_column_fields = ', '.join([f'SOURCE.{column_name}' for column_name in source_diagnostic_column_names])

        join_condition = " AND ".join(
            [
                f"SOURCE.{source_column_name} = TARGET.{target_column_names[index]}"
                for index, source_column_name in enumerate(source_column_names)
            ]
        )

        # Search for all rows where:
        # 1. source value is not null - to avoid null values triggering fails
        # 2. target value is null - this means that source value was not found in target column.
        # Passing query is same on source side, but not null on target side.
        inverse = check_cfg.is_reverse
        where_condition = " OR ".join(
            [
                f"(SOURCE.{source_column_name} IS NOT NULL AND TARGET.{target_column_name} IS {'NOT' if inverse else ''} NULL)"
                for source_column_name, target_column_name in zip(source_column_names, target_column_names)
            ]
        )
        passing_where_condition = " AND ".join(
            [
                f"(SOURCE.{source_column_name} IS NOT NULL AND TARGET.{target_column_name} IS {'' if inverse else 'NOT'} NULL)"
                for source_column_name, target_column_name in zip(source_column_names, target_column_names)
            ]
        )

        partition_filter = self.partition.sql_partition_filter
        if partition_filter:
            scan = self.data_source_scan.scan
            resolved_partition_filter = scan.jinja_resolve(definition=partition_filter)
            where_condition = f"{resolved_partition_filter} AND ({where_condition})"
            passing_where_condition = f"{resolved_partition_filter} AND ({passing_where_condition})"

        jinja_resolve = self.data_source_scan.scan.jinja_resolve

        self.sql = jinja_resolve(
            data_source.sql_reference_query(
                data_source.expr_count_all(), source_table_name, target_table_name, join_condition, where_condition
            )
        )

        self.failed_rows_sql = jinja_resolve(
            data_source.sql_reference_query(
                source_diagnostic_column_fields,
                source_table_name,
                target_table_name,
                join_condition,
                where_condition,
                self.samples_limit,
            )
        )

        self.failing_sql = jinja_resolve(
            data_source.sql_reference_query(
                source_diagnostic_column_fields, source_table_name, target_table_name, join_condition, where_condition
            )
        )

        self.passing_sql = jinja_resolve(
            data_source.sql_reference_query(
                source_diagnostic_column_fields,
                source_table_name,
                target_table_name,
                join_condition,
                passing_where_condition,
            )
        )

    def execute(self):
        self.fetchone()
        missing_reference_count = int(self.row[0])
        self.metric.set_value(missing_reference_count)

        if missing_reference_count and self.samples_limit > 0:
            # TODO: Sample Query execute implicitly stores the failed rows file reference in the passed on metric.
            sample_query = SampleQuery(
                self.data_source_scan,
                self.metric,
                "failed_rows",
                self.failed_rows_sql,
            )
            sample_query.execute()



from typing import Callable, Dict


class DerivedFormula:
    def __init__(self, function: Callable, metric_dependencies: Dict[str, "Metric"]):
        self.function: Callable = function
        self.metric_dependencies: Dict[str, "Metric"] = metric_dependencies

    def collect_failed_rows_sample_refs(self):
        if isinstance(self.metric_dependencies, dict):
            return [
                metric.failed_rows_sample_ref
                for metric in self.metric_dependencies.values()
                if metric.failed_rows_sample_ref is not None
            ]
        return []



from soda.execution.query.query import Query
from soda.execution.query.sample_query import SampleQuery


class DuplicatesQuery(Query):
    def __init__(self, partition: "Partition", metric: "Metric"):
        super().__init__(
            data_source_scan=partition.data_source_scan,
            table=partition.table,
            partition=partition,
            column=metric.column,
            unqualified_query_name=f"duplicate_count",
        )
        self.metric = metric

        self.samples_limit = self.metric.samples_limit

        values_filter_clauses = [f"{column_name} IS NOT NULL" for column_name in self.metric.metric_args]
        partition_filter = self.partition.sql_partition_filter
        if partition_filter:
            scan = self.data_source_scan.scan
            resolved_partition_filter = scan.jinja_resolve(definition=partition_filter)
            values_filter_clauses.append(resolved_partition_filter)

        if metric.filter:
            values_filter_clauses.append(metric.filter)

        values_filter = " \n  AND ".join(values_filter_clauses)

        column_names = ", ".join(self.metric.metric_args)

        table_name = self.partition.table.table_name
        exclude_patterns = self.data_source_scan.data_source.get_exclude_column_patterns_for_table(table_name)
        data_source = self.data_source_scan.data_source
        jinja_resolve = self.data_source_scan.scan.jinja_resolve

        self.sql = jinja_resolve(
            data_source.sql_get_duplicates_count(
                column_names,
                table_name,
                values_filter,
            )
        )

        self.failed_rows_sql = jinja_resolve(
            data_source.sql_get_duplicates(
                column_names,
                table_name,
                values_filter,
                self.samples_limit,
            )
        )
        self.failing_sql = jinja_resolve(
            data_source.sql_get_duplicates(
                column_names,
                table_name,
                values_filter,
                None,
            )
        )

        self.passing_sql = jinja_resolve(
            data_source.sql_get_duplicates(
                column_names,
                table_name,
                values_filter,
                None,
                invert_condition=True,
            )
        )

    def execute(self):
        self.fetchone()
        if self.row:
            duplicates_count = self.row[0]
            self.metric.set_value(duplicates_count)

            if duplicates_count and self.samples_limit > 0:
                # TODO: Sample Query execute implicitly stores the failed rows file reference in the passed on metric.
                sample_query = SampleQuery(
                    self.data_source_scan,
                    self.metric,
                    "failed_rows",
                    self.failed_rows_sql,
                )
                sample_query.execute()

            # TODO: This should be a second failed rows file, refactor failed rows to support multiple files.
            # TODO: removing for now, this is not using standard Query.store() and "gatekeeper" does not kick in - which is a potential data leak.
            # if self.failing_rows_sql_aggregated and self.samples_limit > 0:
            #     aggregate_sample_query = Query(
            #         self.data_source_scan,
            #         self.partition.table,
            #         self.partition,
            #         unqualified_query_name=f"duplicate_count[{'-'.join(self.metric.metric_args)}].failed_rows.aggregated",
            #         sql=self.failing_rows_sql_aggregated,
            #         samples_limit=self.samples_limit,
            #     )
            #     aggregate_sample_query.execute()
            #     self.aggregated_failed_rows_data = aggregate_sample_query.rows



from __future__ import annotations

import collections
import os
from abc import ABC

from soda.cloud.cloud import Cloud
from soda.cloud.historic_descriptor import HistoricDescriptor
from soda.common.attributes_handler import AttributeHandler
from soda.common.string_helper import strip_quotes
from soda.execution.check_outcome import CheckOutcome
from soda.execution.check_type import CheckType
from soda.execution.column import Column
from soda.execution.identity import ConsistentHashBuilder
from soda.execution.metric.metric import Metric
from soda.execution.query.query import Query
from soda.sampler.sample_ref import SampleRef
from soda.sodacl.anomaly_detection_metric_check_cfg import (
    AnomalyDetectionMetricCheckCfg,
)
from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.distribution_check_cfg import DistributionCheckCfg
from soda.sodacl.group_by_check_cfg import GroupByCheckCfg
from soda.sodacl.group_evolution_check_cfg import GroupEvolutionCheckCfg


class Check(ABC):
    @staticmethod
    def create(
        check_cfg: CheckCfg,
        partition: Partition | None = None,
        column: Column | None = None,
        data_source_scan: DataSourceScan | None = None,
    ) -> Check | None:
        from soda.sodacl.anomaly_detection_metric_check_cfg import (
            AnomalyDetectionMetricCheckCfg,
        )
        from soda.sodacl.anomaly_metric_check_cfg import AnomalyMetricCheckCfg
        from soda.sodacl.change_over_time_metric_check_cfg import (
            ChangeOverTimeMetricCheckCfg,
        )
        from soda.sodacl.freshness_check_cfg import FreshnessCheckCfg
        from soda.sodacl.metric_check_cfg import MetricCheckCfg
        from soda.sodacl.reference_check_cfg import ReferenceCheckCfg
        from soda.sodacl.row_count_comparison_check_cfg import (
            RowCountComparisonCheckCfg,
        )
        from soda.sodacl.schema_check_cfg import SchemaCheckCfg
        from soda.sodacl.user_defined_failed_rows_check_cfg import (
            UserDefinedFailedRowsCheckCfg,
        )
        from soda.sodacl.user_defined_failed_rows_expression_check_cfg import (
            UserDefinedFailedRowsExpressionCheckCfg,
        )

        if isinstance(check_cfg, ChangeOverTimeMetricCheckCfg):
            from soda.execution.check.change_over_time_metric_check import (
                ChangeOverTimeMetricCheck,
            )

            return ChangeOverTimeMetricCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, AnomalyMetricCheckCfg):
            from soda.execution.check.anomaly_metric_check import AnomalyMetricCheck

            return AnomalyMetricCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, AnomalyDetectionMetricCheckCfg):
            from soda.execution.check.anomaly_detection_metric_check import (
                AnomalyDetectionMetricCheck,
            )

            return AnomalyDetectionMetricCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, MetricCheckCfg):
            from soda.execution.check.metric_check import MetricCheck

            return MetricCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, SchemaCheckCfg):
            from soda.execution.check.schema_check import SchemaCheck

            return SchemaCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, UserDefinedFailedRowsExpressionCheckCfg):
            from soda.execution.check.user_defined_failed_rows_expression_check import (
                UserDefinedFailedRowsExpressionCheck,
            )

            return UserDefinedFailedRowsExpressionCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, UserDefinedFailedRowsCheckCfg):
            from soda.execution.check.user_defined_failed_rows_check import (
                UserDefinedFailedRowsCheck,
            )

            return UserDefinedFailedRowsCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, RowCountComparisonCheckCfg):
            from soda.execution.check.row_count_comparison_check import (
                RowCountComparisonCheck,
            )

            return RowCountComparisonCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, ReferenceCheckCfg):
            from soda.execution.check.reference_check import ReferenceCheck

            return ReferenceCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, FreshnessCheckCfg):
            from soda.execution.check.freshness_check import FreshnessCheck

            return FreshnessCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, DistributionCheckCfg):
            from soda.execution.check.distribution_check import DistributionCheck

            return DistributionCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, GroupByCheckCfg):
            from soda.execution.check.group_by_check import GroupByCheck

            return GroupByCheck(check_cfg, data_source_scan, partition)

        elif isinstance(check_cfg, GroupEvolutionCheckCfg):
            from soda.execution.check.group_evolution_check import GroupEvolutionCheck

            return GroupEvolutionCheck(check_cfg, data_source_scan, partition)

        raise RuntimeError(f"Bug: Unsupported check type {type(check_cfg)}")

    def __init__(
        self,
        check_cfg: CheckCfg,
        data_source_scan: DataSourceScan,
        partition: Partition | None,
        column: Column | None,
    ):
        from soda.execution.partition import Partition

        self.check_cfg: CheckCfg = check_cfg
        self.logs = data_source_scan.scan._logs
        self.data_source_scan: DataSourceScan = data_source_scan
        self.partition: Partition = partition
        self.column: Column = column
        self.metrics: dict[str, Metric] = {}
        self.attributes = {}
        self.historic_descriptors: dict[str, HistoricDescriptor] = {}
        self.cloud_check_type = "generic"
        # in the evaluate method, checks can optionally extract a failed rows sample ref from the metric
        self.failed_rows_sample_ref: SampleRef | None = None

        # Attribute for automated monitoring
        self.archetype = None

        # Check evaluation outcome
        self.outcome: CheckOutcome | None = None

        # Check outcome reasons in case of fail or pass
        self.outcome_reasons: list[dict] = []
        self.force_send_results_to_cloud = False

        # Default check type is Cloud, when this is set to CheckType.LOCAL, the check will not be sent to cloud
        self.check_type = CheckType.CLOUD

        self.cloud_dict = {}
        self.dict = {}

    @property
    def name(self) -> str:
        """User readable name.

        Uses user provided name if available or generates one from the check definition and thresholds.
        """
        jinja_resolve = self.data_source_scan.scan.jinja_resolve
        if self.check_cfg.name:
            return jinja_resolve(self.check_cfg.name)

        name = self.check_cfg.source_line

        if self.check_cfg.source_configurations:
            source_cfg = dict(self.check_cfg.source_configurations)

            if source_cfg.get("warn"):
                name += f" warn {source_cfg['warn']}"

            if source_cfg.get("fail"):
                name += f" fail {source_cfg['fail']}"

        return jinja_resolve(name)

    @property
    def is_deprecated(self) -> bool:
        return False

    def create_definition(self) -> str:
        check_cfg: CheckCfg = self.check_cfg
        from soda.common.yaml_helper import to_yaml_str

        if isinstance(check_cfg.source_configurations, dict):
            return to_yaml_str({check_cfg.source_header: [{check_cfg.source_line: check_cfg.source_configurations}]})
        else:
            return f"{check_cfg.source_header}:\n  {check_cfg.source_line}"

    def create_identity(self, with_datasource: bool | str = False, with_filename: bool = False) -> str:
        check_cfg: CheckCfg = self.check_cfg
        from soda.common.yaml_helper import to_yaml_str

        hash_builder = ConsistentHashBuilder()
        # Note: In case of for each table, the check_cfg.source_header will contain the actual table name as well
        hash_builder.add(check_cfg.source_header)
        hash_builder.add(check_cfg.source_line)

        # If the check is automated monitoring append "automated_monitoring" tag
        if hasattr(check_cfg, "is_automated_monitoring") and check_cfg.is_automated_monitoring:
            hash_builder.add("automated_monitoring")

        if isinstance(check_cfg.source_configurations, dict):
            identity_source_configurations = dict(check_cfg.source_configurations)
            # The next lines ensures that some of the configuration properties are ignored for computing the check identity
            identity_source_configurations.pop("name", None)
            identity_source_configurations.pop("samples limit", None)
            identity_source_configurations.pop("identity", None)
            identity_source_configurations.pop("attributes", None)
            identity_source_configurations.pop("template", None)
            identity_source_configurations.pop("warn_only", None)

            # Exlude hyperparameters / tuning configurations from identity for anomaly detection checks
            if isinstance(check_cfg, AnomalyDetectionMetricCheckCfg):
                identity_source_configurations.pop("take_over_existing_anomaly_score_check", None)
                identity_source_configurations.pop("training_dataset_parameters", None)
                identity_source_configurations.pop("model", None)
                identity_source_configurations.pop("severity_level_parameters", None)

            if len(identity_source_configurations) > 0:
                # The next line ensures that ordering of the check configurations don't matter for identity
                identity_source_configurations = collections.OrderedDict(sorted(identity_source_configurations.items()))
                identity_source_configurations_yaml = to_yaml_str(identity_source_configurations)
                hash_builder.add(identity_source_configurations_yaml)
        # Temp solution to introduce new variant of identity to help cloud identifying datasets with same name
        # See https://sodadata.atlassian.net/browse/CLOUD-1143
        if with_datasource:
            # Temp workaround to provide migration identities with fixed data source
            # name. See https://sodadata.atlassian.net/browse/CLOUD-5446
            for identity in self.identity_datasource_part() if isinstance(with_datasource, bool) else [with_datasource]:
                hash_builder.add(identity)

        if with_filename:
            hash_builder.add(os.path.basename(self.check_cfg.location.file_path))

        return hash_builder.get_hash()

    # Migrate Identities are created specifically to resolve https://sodadata.atlassian.net/browse/CLOUD-5447?focusedCommentId=30022
    # and can eventually be removed when all checks are migrated.
    def create_migrate_identities(self):
        migrate_data_source_name = self.data_source_scan.data_source.migrate_data_source_name
        if (
            migrate_data_source_name is None
            or self.data_source_scan.data_source.data_source_name == migrate_data_source_name
        ):
            return None

        identities = {
            "v1": self.create_identity(with_datasource=False, with_filename=False),
            "v2": self.create_identity(with_datasource=migrate_data_source_name, with_filename=False),
            "v3": self.create_identity(with_datasource=migrate_data_source_name, with_filename=True),
        }
        if isinstance(self.check_cfg.source_configurations, dict):
            identity = self.check_cfg.source_configurations.get("identity")
            if isinstance(identity, str):
                # append custom identity latest
                identities[f"v{len(identities) + 1}"] = identity
        return identities

    def identity_datasource_part(self) -> list[str]:
        return [
            self.data_source_scan.data_source.data_source_name,
        ]

    def add_outcome_reason(self, outcome_type: str, message: str, severity: str):
        self.force_send_results_to_cloud = True
        self.outcome_reasons.append({"code": outcome_type, "message": message, "severity": severity})  # error/warn/info

    @staticmethod
    def __check_source_to_yaml(source_header: str, source_line: str, source_configurations: dict | None) -> str:
        from soda.common.yaml_helper import to_yaml_str

        if not isinstance(source_configurations, dict):
            return f"{source_header}:\n  {source_line}"
        return to_yaml_str({source_header: [{source_line: source_configurations}]})

    def create_identities(self):
        identities = {
            "v1": self.create_identity(with_datasource=False, with_filename=False),
            "v2": self.create_identity(with_datasource=True, with_filename=False),
            "v3": self.create_identity(with_datasource=True, with_filename=True),
            # v4 is reserved for custom identity
        }
        if isinstance(self.check_cfg.source_configurations, dict):
            identity = self.check_cfg.source_configurations.get("identity")
            if isinstance(identity, str):
                identities["v4"] = identity
        return identities

    # Migrate Identities are created specifically to resolve https://sodadata.atlassian.net/browse/CLOUD-5447?focusedCommentId=30022
    # and can eventually be removed when all checks are migrated.
    def create_migrate_identities(self) -> dict[str, str] | None:
        migrate_data_source_name = self.data_source_scan.data_source.migrate_data_source_name
        if (
            migrate_data_source_name is None
            or self.data_source_scan.data_source.data_source_name == migrate_data_source_name
        ):
            return None

        identities = {
            "v1": self.create_identity(with_datasource=False, with_filename=False),
            "v2": self.create_identity(with_datasource=migrate_data_source_name, with_filename=False),
            "v3": self.create_identity(with_datasource=migrate_data_source_name, with_filename=True),
            # v4 is reserved for custom identity
        }
        if isinstance(self.check_cfg.source_configurations, dict):
            identity = self.check_cfg.source_configurations.get("identity")
            if isinstance(identity, str):
                identities["v4"] = identity
        return identities

    def get_cloud_dict(self):
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        self.cloud_dict.update(
            {
                # See https://sodadata.atlassian.net/browse/CLOUD-1143
                "identity": self.create_identity(with_datasource=True, with_filename=True),
                "identities": self.create_identities(),
                "migratedIdentities": self.create_migrate_identities(),
                "name": self.name,
                "type": self.cloud_check_type,
                "definition": self.create_definition(),
                "resourceAttributes": self._format_attributes(),
                "location": self.check_cfg.location.get_cloud_dict(),
                "dataSource": self.data_source_scan.data_source.data_source_name,
                "table": strip_quotes(Partition.get_table_name(self.partition)),
                # "filter": Partition.get_partition_name(self.partition), TODO: re-enable once backend supports the property.
                "column": Column.get_partition_name(self.column),
                "metrics": [metric.identity for metric in self.metrics.values()],
                "outcome": self.outcome.value if self.outcome else None,
                "diagnostics": self.get_cloud_diagnostics_dict(),
                "source": "soda-core",
            }
        )
        # Update dict if automated monitoring is running
        if self.archetype is not None:
            self.cloud_dict.update({"archetype": self.archetype})

        # Update dict if check is skipped and we want to push reason to cloud
        if self.outcome_reasons:
            self.cloud_dict.update({"outcomeReasons": self.outcome_reasons})
        return self.cloud_dict

    def get_dict(self):
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        self.dict.update(
            {
                "identity": self.create_identity(with_datasource=True, with_filename=True),
                "name": self.name,
                "type": self.cloud_check_type,
                "definition": self.create_definition(),
                "resourceAttributes": self._format_attributes(),
                "location": self.check_cfg.location.get_dict(),
                "dataSource": self.data_source_scan.data_source.data_source_name,
                "table": strip_quotes(Partition.get_table_name(self.partition)),
                "filter": Partition.get_partition_name(self.partition),
                "column": Column.get_partition_name(self.column),
                "metrics": [metric.identity for metric in self.metrics.values()],
                "outcome": self.outcome.value if self.outcome else None,
                "outcomeReasons": self.outcome_reasons,
                "archetype": self.archetype,
                "diagnostics": self.get_cloud_diagnostics_dict(),
            }
        )

        # "contract check id" is a property used by contracts implementation.
        # Here we propagate it from the check source configuration to the check result so that
        # the contracts implementation can correlate the sodacl check result with the contract check
        if self.check_cfg.source_configurations and "identity" in self.check_cfg.source_configurations:
            self.dict["source_identity"] = self.check_cfg.source_configurations.get("identity")

        return self.dict

    def get_cloud_diagnostics_dict(self) -> dict:
        csv_max_length = Cloud.CSV_TEXT_MAX_LENGTH

        cloud_diagnostics = {
            "blocks": [],
            "value": self.check_value if hasattr(self, "check_value") else None,
        }

        if self.failed_rows_sample_ref:
            if self.cloud_check_type == "generic":
                queries = self._get_all_related_queries()
                has_analysis_block = False
                sample_ref_block = self.failed_rows_sample_ref.get_cloud_diagnostics_block()

                for query in queries:
                    if query.failing_sql and query.passing_sql and sample_ref_block:
                        has_analysis_block = True
                        total_failing_rows = self.check_value

                        # TODO: temporary workaround until derived checks are refactored.
                        derived_failing_rows_formula_values = ["invalid_count", "duplicate_count", "missing_count"]
                        if hasattr(self, "formula_values") and self.formula_values:
                            for derived_value in derived_failing_rows_formula_values:
                                if derived_value in self.formula_values:
                                    total_failing_rows = self.formula_values[derived_value]

                        rca_block = {
                            "type": "failedRowsAnalysis",
                            "title": "Failed Rows Analysis",
                            "file": sample_ref_block["file"],
                            "failingRowsQueryName": f"{query.query_name}.failing_sql",
                            "passingRowsQueryName": f"{query.query_name}.passing_sql",
                            "totalFailingRows": total_failing_rows,
                            "sampleRowCount": sample_ref_block["file"]["storedRowCount"],
                        }
                        cloud_diagnostics["blocks"].append(rca_block)

                    # TODO: This should be a second failed rows file, refactor failed rows to support multiple files.
                    if (
                        query.failing_rows_sql_aggregated
                        and hasattr(query, "aggregated_failed_rows_data")
                        and query.aggregated_failed_rows_data
                    ):
                        text = f'{",".join(query.metric.metric_args)},frequency'

                        for row in query.aggregated_failed_rows_data:
                            row_str = f'\n{",".join(map(str, row))}'
                            if len(text) + len(row_str) < csv_max_length:
                                text += row_str

                        aggregate_rows_block = {
                            "type": "csv",
                            "title": "Failed Rows Aggregate",
                            "text": text,
                        }
                        cloud_diagnostics["blocks"].append(aggregate_rows_block)

                if not has_analysis_block:
                    cloud_diagnostics["blocks"].append(self.failed_rows_sample_ref.get_cloud_diagnostics_block())

            else:
                cloud_diagnostics["blocks"].append(self.failed_rows_sample_ref.get_cloud_diagnostics_block())

        return cloud_diagnostics

    def evaluate(self, metrics: dict[str, Metric], historic_values: dict[str, object]):
        raise NotImplementedError("Implement this abstract method")

    def get_log_diagnostic_lines(self) -> list[str]:
        log_diagnostic_lines = []

        log_diagnostic_dict = self.get_log_diagnostic_dict()
        if log_diagnostic_dict:
            for key, value in log_diagnostic_dict.items():
                log_diagnostic_lines.append(f"{key}: {value}")

        return log_diagnostic_lines

    def get_log_diagnostic_dict(self) -> dict:
        diagnostics_dict = self.get_cloud_diagnostics_dict()
        diagnostics_dict.pop("failedRowsFile", None)
        diagnostics_dict.pop("blocks", None)
        return diagnostics_dict

    def __str__(self):
        log_diagnostic_lines = self.get_log_diagnostic_lines()
        if log_diagnostic_lines:
            diagnostics_text = ", ".join(log_diagnostic_lines)
            return f"[{self.check_cfg.source_line}] {self.outcome} ({diagnostics_text})"
        else:
            return f"[{self.check_cfg.source_line}] {self.outcome}"

    def _format_attributes(
        self,
    ) -> list[dict[str, any]]:
        attribute_handler = AttributeHandler(self.logs)

        attributes = []
        for k, v in self.attributes.items():
            attributes.append(
                {
                    "name": k,
                    "value": attribute_handler.format_attribute(v),
                }
            )

        return attributes

    def _get_all_related_queries(self) -> list(Query):
        queries = []

        for metric in self.metrics.values():
            for query in metric.queries:
                queries.append(query)

        return queries



from typing import Dict

from soda.execution.check.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.metric.metric import Metric
from soda.execution.metric.reference_metric import ReferenceMetric

KEY_INVALID_REFERENCE_COUNT = "invalid_reference_count"


class ReferenceCheck(Check):
    def __init__(self, check_cfg: "ReferenceCheckCfg", data_source_scan: "DataSourceScan", partition: "Partition"):
        single_source_column_name = (
            check_cfg.source_column_names[0] if len(check_cfg.source_column_names) == 1 else None
        )
        single_source_column = (
            partition.table.get_or_create_column(single_source_column_name) if single_source_column_name else None
        )
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=single_source_column,
        )
        metric = ReferenceMetric(
            data_source_scan=self.data_source_scan,
            check=self,
            partition=partition,
            single_source_column=single_source_column,
        )
        metric = self.data_source_scan.resolve_metric(metric)
        self.metrics[KEY_INVALID_REFERENCE_COUNT] = metric
        self.failed_rows_sample_ref = None

    def get_cloud_diagnostics_dict(self) -> dict:
        cloud_diagnostics = super().get_cloud_diagnostics_dict()

        return cloud_diagnostics

    def evaluate(self, metrics: Dict[str, Metric], historic_values: Dict[str, object]):
        metric = metrics.get(KEY_INVALID_REFERENCE_COUNT)
        invalid_reference_count: int = metric.value

        self.outcome = CheckOutcome.PASS
        if invalid_reference_count > 0:
            self.outcome = CheckOutcome.FAIL

        self.check_value = invalid_reference_count
        self.failed_rows_sample_ref = metric.failed_rows_sample_ref



from __future__ import annotations

from numbers import Number

from ruamel.yaml import YAML
from soda.cli.cli import DATA_SOURCES_WITH_DISTRIBUTION_CHECK_SUPPORT
from soda.common.exceptions import SODA_SCIENTIFIC_MISSING_LOG_MESSAGE
from soda.execution.check.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.column import Column
from soda.execution.data_source_scan import DataSourceScan
from soda.execution.metric.metric import Metric
from soda.execution.partition import Partition
from soda.execution.query.query import Query
from soda.sodacl.distribution_check_cfg import DistributionCheckCfg


class DistributionCheck(Check):
    def __init__(
        self,
        check_cfg: DistributionCheckCfg,
        data_source_scan: DataSourceScan,
        partition: Partition | None = None,
        column: Column | None = None,
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
        )

        if data_source_scan.data_source.type not in DATA_SOURCES_WITH_DISTRIBUTION_CHECK_SUPPORT:
            self.logs.info(
                f"The support for your data source type is experimental. Distribution checks are"
                f" not tested for '{data_source_scan.data_source.type}' and may not work."
            )

        self.distribution_check_cfg: DistributionCheckCfg = self.check_cfg
        metric = Metric(
            data_source_scan=self.data_source_scan,
            name=self.distribution_check_cfg.source_line,
            partition=partition,
            column=column,
            check=self,
            identity_parts=["distribution"],
        )
        metric.value = None
        metric = data_source_scan.resolve_metric(metric)
        self.metrics["distribution-difference-metric"] = metric
        self.check_value: float | None = None
        self.parsed_dro = dict(
            YAML().load(
                self.data_source_scan.scan._read_file(
                    file_type="disribution reference object yaml",
                    file_path=self.distribution_check_cfg.reference_file_path,
                )
            )
        )
        self.distribution_name = self.distribution_check_cfg.distribution_name
        self.distribution_type = self.get_distribution_type()
        self.max_limit = int(1e6)

    def evaluate(self, metrics: dict[str, Metric], historic_values: dict[str, object]) -> None:
        try:
            from soda.scientific.common.exceptions import LoggableException
            from soda.scientific.distribution.comparison import DistributionChecker
        except ModuleNotFoundError as e:
            self.logs.error(f"{SODA_SCIENTIFIC_MISSING_LOG_MESSAGE}\n Original error: {e}")
            return

        sql = self.sql_column_values_query(self.distribution_check_cfg)
        self.logs.debug(f"Executing query for the distribution check: \n{sql}")

        self.query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name="get_values_for_distro_check",
            sql=sql,
        )
        self.query.execute()
        if self.query.exception is None and self.query.rows is not None:
            ref_file_path = self.distribution_check_cfg.reference_file_path
            dist_method = self.distribution_check_cfg.method
            dist_name = self.distribution_check_cfg.distribution_name
            try:
                if self.distribution_type == "categorical":
                    # Collect test data as a list of tuples (value, count)
                    test_data = self.query.rows
                else:
                    # Collect test data as a list of values
                    test_data = [row[0] for row in self.query.rows]

                    # If all values are null, we should not run the check
                    if len(test_data) > 0 and all(value is None for value in test_data):
                        self.logs.warning(
                            f"All values are null in your test data. Skipping distribution check for column '{self.column.column_name}'"
                        )
                        return
                check_result_dict = DistributionChecker(
                    dist_method=dist_method,
                    parsed_dro=self.parsed_dro,
                    dist_ref_file_path=ref_file_path,
                    dist_name=dist_name,
                    data=test_data,
                    max_limit=self.max_limit,
                    logs=self.logs,
                ).run()
                self.check_value = check_result_dict["check_value"]
                self.metrics["distribution-difference-metric"].value = self.check_value
                self.set_outcome_based_on_check_value()
            except LoggableException as e:
                self.logs.error(e, location=self.check_cfg.location)

    def set_outcome_based_on_check_value(self) -> None:
        if self.check_value is not None and (
            self.distribution_check_cfg.warn_threshold_cfg or self.distribution_check_cfg.fail_threshold_cfg
        ):
            if isinstance(self.check_value, Number):
                if (
                    self.distribution_check_cfg.fail_threshold_cfg
                    and self.distribution_check_cfg.fail_threshold_cfg.is_bad(self.check_value)
                ):
                    self.outcome = CheckOutcome.FAIL
                elif (
                    self.distribution_check_cfg.warn_threshold_cfg
                    and self.distribution_check_cfg.warn_threshold_cfg.is_bad(self.check_value)
                ):
                    self.outcome = CheckOutcome.WARN
                else:
                    self.outcome = CheckOutcome.PASS
            else:
                self.logs.error(
                    f"Cannot evaluate check: Expected a numeric value, but was {self.check_value}",
                    location=self.check_cfg.location,
                )

    def get_cloud_diagnostics_dict(self) -> dict:
        distribution_check_cfg: DistributionCheckCfg = self.check_cfg
        cloud_diagnostics = {"value": self.check_value}
        if distribution_check_cfg.fail_threshold_cfg is not None:
            cloud_diagnostics["fail"] = distribution_check_cfg.fail_threshold_cfg.to_soda_cloud_diagnostics_json()
        if distribution_check_cfg.warn_threshold_cfg is not None:
            cloud_diagnostics["warn"] = distribution_check_cfg.warn_threshold_cfg.to_soda_cloud_diagnostics_json()
        return cloud_diagnostics

    def get_log_diagnostic_dict(self) -> dict:
        log_diagnostics = super().get_log_diagnostic_dict()
        # if self.historic_diff_values:
        #     log_diagnostics.update(self.historic_diff_values)
        return log_diagnostics

    def get_distribution_type(self) -> str:
        dist_name = self.distribution_check_cfg.distribution_name
        if dist_name is None:
            dist_type = self.parsed_dro["distribution_type"]
        else:
            dist_type = self.parsed_dro[dist_name]["distribution_type"]
        return dist_type

    def sql_column_values_query(self, distribution_check_cfg: DistributionCheckCfg) -> str:
        column_name = distribution_check_cfg.column_name
        scan = self.data_source_scan.scan

        partition_filter = scan.jinja_resolve(self.partition.sql_partition_filter)
        distribution_check_filter = scan.jinja_resolve(distribution_check_cfg.filter)
        filters = []
        filters.append(partition_filter)
        filters.append(distribution_check_filter)

        filter_clause = " AND ".join(_filter for _filter in filters if _filter)
        sample_clause = None
        limit = None

        if self.distribution_type == "continuous":
            sample_clause = scan.jinja_resolve(distribution_check_cfg.sample_clause)
            if sample_clause is not None:
                limit = None  # No need to apply limit if we are sampling
            else:
                limit = self.max_limit
        sql = self.data_source_scan.data_source.sql_select_column_with_filter_and_limit(
            column_name=column_name,
            table_name=self.partition.table.qualified_table_name,
            filter_clause=filter_clause,
            sample_clause=sample_clause,
            limit=limit,
        )
        if self.distribution_type == "categorical":
            sql = self.data_source_scan.data_source.sql_groupby_count_categorical_column(
                select_query=sql,
                column_name=column_name,
                limit=self.max_limit,
            )
        return sql

    def get_summary(self) -> str:
        error_summary = (
            " NO THRESHOLD PROVIDED ->"
            if not self.check_cfg.fail_threshold_cfg and not self.check_cfg.warn_threshold_cfg
            else ""
        )
        summary_message = f"{super().get_summary()}{error_summary}"
        return summary_message



ERROR_CODE_GENERIC = "generic_error"
ERROR_CODE_CONNECTION_FAILED = "connection_failed"
SODA_SCIENTIFIC_MISSING_LOG_MESSAGE = (
    "The anomaly detection module could not be imported. "
    "This is often the case when the soda-scientific package was not installed. "
    "\n Please check https://docs.soda.io/soda-core/installation.html#install-soda-core-scientific "
    "for instructions."
)


class SodaSqlError(Exception):
    def __init__(self, msg, original_exception):
        super().__init__(f"{msg}: {str(original_exception)}")
        self.error_code = ERROR_CODE_GENERIC
        self.original_exception = original_exception


class DataSourceError(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.error_code = ERROR_CODE_CONNECTION_FAILED


class DataSourceConnectionError(SodaSqlError):
    def __init__(self, data_source_type, original_exception):
        super().__init__(
            f"Encountered a problem while trying to connect to {data_source_type}",
            original_exception,
        )
        self.error_code = ERROR_CODE_CONNECTION_FAILED
        self.data_source_type = data_source_type



from typing import Dict, Optional

from soda.cloud.historic_descriptor import HistoricChangeOverTimeDescriptor
from soda.execution.check.metric_check import MetricCheck
from soda.execution.metric.metric import Metric
from soda.sodacl.metric_check_cfg import MetricCheckCfg

KEY_HISTORIC_METRIC_AGGREGATE = "historic_metric_aggregate"


class ChangeOverTimeMetricCheck(MetricCheck):
    def __init__(
        self,
        check_cfg: "MetricCheckCfg",
        data_source_scan: "DataSourceScan",
        partition: Optional["Partition"] = None,
        column: Optional["Column"] = None,
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
        )
        metric_check_cfg: MetricCheckCfg = self.check_cfg
        metric_name = metric_check_cfg.metric_name
        metric = self.metrics[metric_name]

        self.historic_descriptors[KEY_HISTORIC_METRIC_AGGREGATE] = HistoricChangeOverTimeDescriptor(
            metric_identity=metric.identity,
            change_over_time_cfg=metric_check_cfg.change_over_time_cfg,
        )

    def evaluate(self, metrics: Dict[str, Metric], historic_values: Dict[str, object]):
        metric_value = self.get_metric_value()

        historic_results = historic_values.get(KEY_HISTORIC_METRIC_AGGREGATE).get("measurements").get("results")

        if historic_results and historic_results[0].get("value") is not None:
            historic_value = historic_results[0].get("value")
            if isinstance(metric_value, int) and isinstance(historic_value, int):
                self.check_value = metric_value - historic_value
            else:
                self.check_value = float(metric_value) - float(historic_value)

            historic_descriptor = self.historic_descriptors[KEY_HISTORIC_METRIC_AGGREGATE]
            if isinstance(historic_descriptor, HistoricChangeOverTimeDescriptor):
                if historic_descriptor.change_over_time_cfg.percent:
                    if historic_value == 0:
                        if self.check_value == 0:
                            self.check_value = 0
                        else:
                            self.check_value = None
                    else:
                        self.check_value = self.check_value / historic_value * 100

            self.historic_diff_values = {
                "historic_value": historic_value,
                "metric_value": metric_value,
            }

            self.set_outcome_based_on_check_value()

        else:
            self.logs.info("Skipping metric check eval because there is not enough historic data yet")

    def get_cloud_diagnostics_dict(self) -> dict:
        cloud_diagnostics = super().get_cloud_diagnostics_dict()
        if self.historic_diff_values:
            cloud_diagnostics["diagnostics"] = self.historic_diff_values
        return cloud_diagnostics

    def get_log_diagnostic_dict(self) -> dict:
        log_diagnostics = super().get_log_diagnostic_dict()
        if self.historic_diff_values:
            log_diagnostics.update(self.historic_diff_values)
        return log_diagnostics



from typing import Tuple

from soda.common.memory_safe_cursor_fetcher import MemorySafeCursorFetcher
from soda.sampler.sample import Sample
from soda.sampler.sample_schema import SampleColumn, SampleSchema


class DbSample(Sample):
    def __init__(self, cursor, data_source, limit=None):
        self.cursor = cursor
        self.safe_fetcher = MemorySafeCursorFetcher(cursor)
        self.data_source = data_source
        self.rows = None
        self._limit = limit

    def get_rows(self) -> Tuple[Tuple]:
        return self.safe_fetcher.get_rows()

    def get_rows_count(self) -> int:
        return self.safe_fetcher.get_row_count()

    def get_schema(self) -> SampleSchema:
        return self._convert_python_db_schema_to_sample_schema(self.cursor.description)

    def _convert_python_db_schema_to_sample_schema(self, dbapi_description) -> SampleSchema:
        columns = SampleColumn.create_sample_columns(dbapi_description, self.data_source)
        return SampleSchema(columns=columns)



from dataclasses import dataclass
from datetime import timedelta
from hashlib import blake2b
from numbers import Number
from typing import Optional


class Identity:
    @staticmethod
    def property(field_name, value):
        return IdentityProperty(field_name, value)

    @staticmethod
    def create_identity(
        identity_type: str,
        data_source_scan: "DataSourceScan",
        partition: "Partition",
        column: "Column",
        name: Optional[str],
        is_automated_monitoring: bool,
        identity_parts: list,
    ):
        parts = [identity_type]
        if data_source_scan.scan._scan_definition_name:
            parts.append(data_source_scan.scan._scan_definition_name)
        parts.append(data_source_scan.data_source.data_source_name)

        table = partition.table if partition else None
        if table:
            parts.append(table.table_name)
        if partition and partition.partition_name is not None:
            parts.append(partition.partition_name)
        if column is not None:
            parts.append(column.column_name)
        if name:
            parts.append(name)
        if is_automated_monitoring:
            parts.append("automated_monitoring")
        if identity_parts:
            hash_builder = ConsistentHashBuilder()
            for identity_hash_part in identity_parts:
                hash_builder.add(identity_hash_part)
            hash = hash_builder.get_hash()
            if hash:
                parts.append(hash)

        return "-".join([str(p) for p in parts])


@dataclass
class IdentityProperty:
    field_name: str
    value: object


class ConsistentHashBuilder:
    def __init__(self, hash_string_length: int = 8):
        if hash_string_length % 2 != 0:
            raise AssertionError(f"hash_string_length must be divisible by 2: {hash_string_length} is not")
        self.hash_string_length = hash_string_length
        self.blake2b = None

    def get_blake2b(self) -> blake2b:
        # Lazy initialization of blake2b in order to return None in the self.get_hash(self) in case nothing was added
        if self.blake2b is None:
            self.blake2b = blake2b(digest_size=int(self.hash_string_length / 2))
        return self.blake2b

    def add(self, value: Optional[str]):
        from soda.sodacl.change_over_time_cfg import ChangeOverTimeCfg
        from soda.sodacl.location import Location
        from soda.sodacl.missing_and_valid_cfg import MissingAndValidCfg
        from soda.sodacl.schema_check_cfg import SchemaValidations
        from soda.sodacl.threshold_cfg import ThresholdCfg

        if value is None:
            return
        elif isinstance(value, str):
            self.get_blake2b().update(value.encode("utf-8"))
        elif isinstance(value, Number) or isinstance(value, bool):
            self.get_blake2b().update(str(value).encode("utf-8"))
        elif isinstance(value, list) or isinstance(value, dict):
            self.add_all(value)
        elif isinstance(value, timedelta):
            self.add(str(value))
        elif isinstance(value, IdentityProperty):
            if value.value is not None:
                self.add(value.field_name)
                self.add(value.value)
        elif (
            isinstance(value, Location)
            or isinstance(value, MissingAndValidCfg)
            or isinstance(value, ChangeOverTimeCfg)
            or isinstance(value, ThresholdCfg)
            or isinstance(value, SchemaValidations)
        ):
            self.add_all(value.get_identity_parts())
        else:
            raise AssertionError(f"Expected str, number or None, not {value}")

    def add_all(self, collection):
        if isinstance(collection, list):
            for e in collection:
                self.add(e)
        elif isinstance(collection, dict):
            for k, v in collection.items():
                self.add(k)
                self.add(v)
        elif collection is not None:
            raise AssertionError(f"Expected list, dict or None, not {collection}")

    def get_hash(self) -> str:
        return self.blake2b.hexdigest() if self.blake2b else None



from __future__ import annotations

from soda.execution.query.query import Query


class UserDefinedFailedRowsQuery(Query):
    def __init__(
        self,
        data_source_scan: DataSourceScan,
        metric: FailedRowsQueryMetric,
        partition: Partition,
        location: Location | None = None,
        samples_limit: int | None = None,
    ):
        super().__init__(
            data_source_scan=data_source_scan,
            unqualified_query_name=f"failed_rows[{metric.name}]",
            location=location,
            samples_limit=samples_limit,
            sql=metric.query,
            partition=partition,
        )
        self.metric = metric

    def execute(self):
        self.store()
        if self.sample_ref:
            self.metric.set_value(self.sample_ref.total_row_count)
            self.metric.failed_rows_sample_ref = self.sample_ref



from __future__ import annotations

import logging
import math
import re
from abc import ABC
from typing import TYPE_CHECKING

import requests
from requests import Response
from soda.__version__ import SODA_CORE_VERSION
from soda.cloud.historic_descriptor import HistoricDescriptor
from soda.common.json_helper import JsonHelper
from soda.common.logs import Logs

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from soda.scan import Scan


class Cloud(ABC):
    CSV_TEXT_MAX_LENGTH = math.inf

    def __init__(
        self,
        host: str,
        api_key_id: str,
        api_key_secret: str,
        token: str | None,
        port: str | None,
        logs: Logs,
        scheme: str = "https",
    ):
        # TODO: revisit if this is generic enough to be in core.
        self.host = host
        self.port = f":{port}" if port else ""
        self.scheme = scheme if scheme else "https"
        self.api_url = f"{self.scheme}://{self.host}{self.port}/api"
        self.api_key_id = api_key_id
        self.api_key_secret = api_key_secret
        self.token: str | None = token
        self.headers = {"User-Agent": f"SodaCore/{SODA_CORE_VERSION}"}
        self.logs = logs
        self._organization_configuration = None

    @property
    def organization_configuration(self) -> dict:
        return self._organization_configuration

    @staticmethod
    def build_scan_results(scan) -> dict:
        checks = [
            check.get_dict()
            for check in scan._checks
            if (check.outcome is not None or check.force_send_results_to_cloud is True) and check.archetype is None
        ]
        automated_monitoring_checks = [
            check.get_dict()
            for check in scan._checks
            if (check.outcome is not None or check.force_send_results_to_cloud is True) and check.archetype is not None
        ]

        profiling = [
            profile_table.get_dict()
            for profile_table in scan._profile_columns_result_tables + scan._sample_tables_result_tables
        ]

        query_list = []
        for query in scan._queries:
            query_list += query.get_dicts()

        return JsonHelper.to_jsonnable(  # type: ignore
            {
                "definitionName": scan._scan_definition_name,
                "defaultDataSource": scan._data_source_name,
                "dataTimestamp": scan._data_timestamp,
                # Can be changed by user, this is shown in Cloud as time of a scan.
                "scanStartTimestamp": scan._scan_start_timestamp,  # Actual time when the scan started.
                "scanEndTimestamp": scan._scan_end_timestamp,  # Actual time when scan ended.
                "hasErrors": scan.has_error_logs(),
                "hasWarnings": scan.has_check_warns(),
                "hasFailures": scan.has_check_fails(),
                "metrics": [metric.get_cloud_dict() for metric in scan._metrics],
                # If archetype is not None, it means that check is automated monitoring
                "checks": checks,
                "queries": query_list,
                "automatedMonitoringChecks": automated_monitoring_checks,
                "profiling": profiling,
                "metadata": [
                    discover_tables_result.get_cloud_dict()
                    for discover_tables_result in scan._discover_tables_result_tables
                ],
                "logs": [log.get_dict() for log in scan._logs.logs],
            }
        )

    @staticmethod
    def _serialize_file_upload_value(value):
        if value is None or isinstance(value, str) or isinstance(value, int) or isinstance(value, float):
            return value
        return str(value)

    def upload_sample(
        self, scan: Scan, sample_rows: tuple[tuple], sample_file_name: str, samples_limit: int | None
    ) -> str:
        raise NotImplementedError()

    def _fileify(self, name: str):
        return re.sub(r"\W+", "_", name).lower()

    def get_historic_data(self, historic_descriptor: HistoricDescriptor):
        raise NotImplementedError()

    def is_samples_disabled(self) -> bool:
        return False

    def _execute_query(self, query: dict, query_name: str):
        return self._execute_request("query", query, False, query_name)

    def _execute_command(self, command: dict, command_name: str):
        return self._execute_request("command", command, False, command_name)

    def _execute_request(self, request_type: str, request_body: dict, is_retry: bool, request_name: str):
        from soda.scan import verbose

        try:
            request_body["token"] = self._get_token()
            if verbose:
                logger.debug(f"{JsonHelper.to_json_pretty(request_body)}")
            response = self._http_post(
                url=f"{self.api_url}/{request_type}", headers=self.headers, json=request_body, request_name=request_name
            )
            response_json = response.json()
            if response.status_code == 401 and not is_retry:
                logger.debug("Authentication failed. Probably token expired. Re-authenticating...")
                self.token = None
                response_json = self._execute_request(request_type, request_body, True, request_name)
            elif response.status_code != 200:
                self.logs.error(
                    f"Error while executing Soda Cloud {request_type} response code: {response.status_code}"
                )
                if verbose:
                    self.logs.debug(response.text)
            return response_json
        except Exception as e:
            self.logs.error(f"Error while executing Soda Cloud {request_type}", exception=e)

    def _http_post(self, request_name: str = None, **kwargs) -> Response:
        response = requests.post(**kwargs)

        if request_name:
            trace_id = response.headers.get("X-Soda-Trace-Id")
            if trace_id:
                self.soda_cloud_trace_ids[request_name] = trace_id

        return response

    def _get_token(self):
        raise NotImplementedError()



from soda.execution.metric.query_metric import QueryMetric
from soda.execution.query.group_evolution_query import GroupEvolutionQuery


class GroupEvolutionMetric(QueryMetric):
    def __init__(self, data_source_scan: "DataSourceScan", partition: "Partition", check: "Check", query: str):
        super().__init__(
            data_source_scan=data_source_scan,
            partition=partition,
            column=None,
            name="group evolution",
            check=check,
            identity_parts=[],
        )
        self.query = query

    def ensure_query(self):
        check = next(iter(self.checks), None)
        location = check.check_cfg.location if check else None
        self.data_source_scan.queries.append(
            GroupEvolutionQuery(
                data_source_scan=self.data_source_scan,
                metric=self,
                location=location,
                partition=self.partition,
            )
        )

    def get_cloud_dict(self):
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        return {
            "identity": self.identity,
            "metricName": self.name,
            "dataSourceName": self.data_source_scan.data_source.data_source_name,
            "tableName": Partition.get_table_name(self.partition),
            "partitionName": Partition.get_partition_name(self.partition),
            "columnName": Column.get_partition_name(self.column),
            "value": self.value,
        }

    def get_dict(self):
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        return {
            "identity": self.identity,
            "metricName": self.name,
            "dataSourceName": self.data_source_scan.data_source.data_source_name,
            "tableName": Partition.get_table_name(self.partition),
            "partitionName": Partition.get_partition_name(self.partition),
            "columnName": Column.get_partition_name(self.column),
            "value": self.value,
        }



from typing import List, Optional

from soda.sodacl.location import Location


class DataSourceCheckCfg:
    def __init__(self, data_source_name: str, location: Location):
        self.data_source_name: str = data_source_name
        self.location: Location = location
        self.include_tables: Optional[List[str]] = []
        self.exclude_tables: Optional[List[str]] = []


class AutomatedMonitoringCfg(DataSourceCheckCfg):
    def __init__(self, data_source_name: str, location: Location):
        super().__init__(data_source_name, location)


class ProfileColumnsCfg(DataSourceCheckCfg):
    def __init__(self, data_source_name: str, location: Location):
        super().__init__(data_source_name, location)
        self.include_columns: List[str] = []
        self.exclude_columns: List[str] = []
        # TODO add parsing for this configuration
        self.limit_mins_maxs: int = 5
        # TODO add parsing for this configuration
        self.limit_frequent_values: int = 10


class DiscoverTablesCfg(DataSourceCheckCfg):
    def __init__(self, data_source_name: str, location: Location):
        super().__init__(data_source_name, location)


class SampleTablesCfg(DataSourceCheckCfg):
    def __init__(self, data_source_name: str, location: Location):
        super().__init__(data_source_name, location)



from abc import ABC, abstractmethod

from soda.sampler.sample_context import SampleContext
from soda.sampler.sample_ref import SampleRef

DEFAULT_FAILED_ROWS_SAMPLE_LIMIT = 100


class Sampler(ABC):
    @abstractmethod
    def store_sample(self, sample_context: SampleContext) -> SampleRef:
        pass



from __future__ import annotations

from soda.sodacl.location import Location

CFG_VALID_VALUES = "valid values"
CFG_INVALID_VALUES = "invalid values"
CFG_VALID_FORMAT = "valid format"
CFG_INVALID_FORMAT = "invalid format"
CFG_VALID_REGEX = "valid regex"
CFG_INVALID_REGEX = "invalid regex"
CFG_VALID_MIN = "valid min"
CFG_VALID_MAX = "valid max"
CFG_VALID_MIN_LENGTH = "valid min length"
CFG_VALID_MAX_LENGTH = "valid max length"
CFG_VALID_LENGTH = "valid length"

CFG_MISSING_VALUES = "missing values"
CFG_MISSING_FORMAT = "missing format"
CFG_MISSING_REGEX = "missing regex"

CFG_MISSING_VALID_ALL = [
    CFG_VALID_VALUES,
    CFG_INVALID_VALUES,
    CFG_VALID_FORMAT,
    CFG_VALID_REGEX,
    CFG_INVALID_FORMAT,
    CFG_INVALID_REGEX,
    CFG_VALID_MIN,
    CFG_VALID_MAX,
    CFG_VALID_MIN_LENGTH,
    CFG_VALID_MAX_LENGTH,
    CFG_VALID_LENGTH,
    CFG_MISSING_VALUES,
    CFG_MISSING_FORMAT,
    CFG_MISSING_REGEX,
]


class MissingAndValidCfg:
    def __init__(self):
        self.missing_values: list[str] | None = None
        self.missing_values_location: Location | None = None
        self.missing_format: str | None = None
        self.missing_format_location: Location | None = None
        self.missing_regex: str | None = None
        self.missing_regex_location: Location | None = None
        # TODO
        # self.missing_expr: Optional[str] = None
        self.valid_values: list[str] | None = None
        self.valid_values_location: Location | None = None
        self.invalid_values: list[str] | None = None
        self.invalid_values_location: Location | None = None
        self.valid_format: str | None = None
        self.valid_format_location: Location | None = None
        self.invalid_format: str | None = None
        self.invalid_format_location: Location | None = None
        self.valid_regex: str | None = None
        self.valid_regex_location: Location | None = None
        self.invalid_regex: str | None = None
        self.invalid_regex_location: Location | None = None
        self.valid_length: int | None = None
        self.valid_length_location: Location | None = None
        self.valid_min_length: int | None = None
        self.valid_min_length_location: Location | None = None
        self.valid_max_length: int | None = None
        self.valid_max_length_location: Location | None = None
        self.valid_min: float | None = None
        self.valid_min_location: Location | None = None
        self.valid_max: float | None = None
        self.valid_max_location: Location | None = None
        # TODO
        # self.valid_expr: Optional[str] = None

    def get_identity_parts(self) -> list:
        from soda.execution.identity import Identity

        return [
            Identity.property("missing_values", self.missing_values),
            Identity.property("missing_format", self.missing_format),
            Identity.property("missing_regex", self.missing_regex),
            Identity.property("valid_values", self.valid_values),
            Identity.property("valid_format", self.valid_format),
            Identity.property("valid_regex", self.valid_regex),
            Identity.property("valid_length", self.valid_length),
            Identity.property("valid_min_length", self.valid_min_length),
            Identity.property("valid_max_length", self.valid_max_length),
            Identity.property("valid_min", self.valid_min),
            Identity.property("valid_max", self.valid_max),
        ]

    def is_empty(self) -> bool:
        return all([var is None for var in vars(self).values()])

    @staticmethod
    def merge(
        missing_and_valid_cfg: MissingAndValidCfg, column_configurations_cfg: ColumnConfigurationsCfg
    ) -> MissingAndValidCfg | None:
        if missing_and_valid_cfg is not None and missing_and_valid_cfg.is_empty():
            missing_and_valid_cfg = None
        if column_configurations_cfg is not None and column_configurations_cfg.is_empty():
            column_configurations_cfg = None

        if missing_and_valid_cfg is None and column_configurations_cfg is None:
            return None
        if missing_and_valid_cfg is not None and column_configurations_cfg is None:
            return missing_and_valid_cfg
        if missing_and_valid_cfg is None and column_configurations_cfg is not None:
            missing_and_valid_cfg = MissingAndValidCfg()
            missing_and_valid_cfg.__merge(column_configurations_cfg)
            return missing_and_valid_cfg

        # Merge
        missing_and_valid_cfg.__merge(column_configurations_cfg)
        return missing_and_valid_cfg

    def __merge(self, other: MissingAndValidCfg):
        if other.missing_values is not None:
            if self.missing_values is None:
                self.missing_values = []
                self.missing_values_location = other.missing_values_location
            self.missing_values.extend(other.missing_values)

        if self.missing_format is None and other.missing_format is not None:
            self.missing_format = other.missing_format
            self.missing_format_location = other.missing_format_location

        if self.missing_regex is None and other.missing_regex is not None:
            self.missing_regex = other.missing_regex
            self.missing_regex_location = other.missing_regex_location

        if other.valid_values is not None:
            if self.valid_values is None:
                self.valid_values = []
                self.valid_values_location = other.valid_values_location
            self.valid_values.extend(other.valid_values)

        if self.valid_format is None and other.valid_format is not None:
            self.valid_format = other.valid_format
            self.valid_format_location = other.valid_format_location

        if self.valid_regex is None and other.valid_regex is not None:
            self.valid_regex = other.valid_regex
            self.valid_regex_location = other.valid_regex_location

        if self.valid_length is None and other.valid_length is not None:
            self.valid_length = other.valid_length
            self.valid_length_location = other.valid_length_location

        if self.valid_min_length is None and other.valid_min_length is not None:
            self.valid_min_length = other.valid_min_length
            self.valid_min_length_location = other.valid_min_length_location

        if self.valid_max_length is None and other.valid_max_length is not None:
            self.valid_max_length = other.valid_max_length
            self.valid_max_length_location = other.valid_max_length_location

        if self.valid_min is None and other.valid_min is not None:
            self.valid_min = other.valid_min
            self.valid_min_location = other.valid_min_location

        if self.valid_max is None and other.valid_max is not None:
            self.valid_max = other.valid_max
            self.valid_max_location = other.valid_max_location



from __future__ import annotations

from datetime import timezone

from soda.cloud.historic_descriptor import (
    HistoricCheckResultsDescriptor,
    HistoricMeasurementsDescriptor,
)
from soda.common.exceptions import SODA_SCIENTIFIC_MISSING_LOG_MESSAGE
from soda.execution.check.metric_check import MetricCheck
from soda.execution.check_outcome import CheckOutcome
from soda.execution.column import Column
from soda.execution.data_source_scan import DataSourceScan
from soda.execution.metric.metric import Metric
from soda.execution.partition import Partition
from soda.sodacl.metric_check_cfg import MetricCheckCfg

KEY_HISTORIC_MEASUREMENTS = "historic_measurements"
KEY_HISTORIC_CHECK_RESULTS = "historic_check_results"
HISTORIC_MEASUREMENTS_LIMIT = 1000


class AnomalyMetricCheck(MetricCheck):
    def __init__(
        self,
        check_cfg: MetricCheckCfg,
        data_source_scan: DataSourceScan,
        partition: Partition | None = None,
        column: Column | None = None,
    ):
        try:
            super().__init__(
                check_cfg=check_cfg,
                data_source_scan=data_source_scan,
                partition=partition,
                column=column,
            )
            self.skip_anomaly_check = False
            metric_check_cfg: MetricCheckCfg = self.check_cfg
            metric_name = metric_check_cfg.metric_name
            metric = self.metrics[metric_name]

            if not metric_check_cfg.fail_threshold_cfg and not metric_check_cfg.warn_threshold_cfg:
                self.skip_anomaly_check = True
                self.logs.error(
                    "Non default threshold for anomaly detection is currently not supported. "
                    "Make sure that your check definition adheres to the following "
                    f"syntax: 'anomaly score for {metric} < default'."
                )

            self.historic_descriptors[KEY_HISTORIC_MEASUREMENTS] = HistoricMeasurementsDescriptor(
                metric_identity=metric.identity,
                limit=HISTORIC_MEASUREMENTS_LIMIT,
            )
            self.historic_descriptors[KEY_HISTORIC_CHECK_RESULTS] = HistoricCheckResultsDescriptor(
                check_identity=self.create_identity(), limit=HISTORIC_MEASUREMENTS_LIMIT
            )
            self.diagnostics = {}
            self.cloud_check_type = "anomalyDetection"
        except Exception as e:
            self.skip_anomaly_check = True
            data_source_scan.scan._logs.error(
                f"""An error occurred during the initialization of AnomalyMetricCheck. Please make sure"""
                f""" that the metric '{check_cfg.metric_name}' is supported. For more information see"""
                f""" the docs: https://docs.soda.io/soda-cl/anomaly-score.html#anomaly-score-checks.""",
                exception=e,
            )

    def evaluate(self, metrics: dict[str, Metric], historic_values: dict[str, object]) -> None:
        if self.skip_anomaly_check:
            return

        self.logs.info(
            "Anomaly Score Deprecation Warning: The anomaly score check is deprecated "
            "and will be removed in the future. Please consider using the anomaly "
            "detection check going forward. See the migration guide: "
            "https://docs.soda.io/soda-cl/anomaly-detection#migrate-to-anomaly-detection"
        )
        metric_name = self.check_cfg.metric_name
        # check that we get data objects from cloud that we can work with
        if isinstance(historic_values, dict):
            historic_measurements = historic_values.get(KEY_HISTORIC_MEASUREMENTS, {}).get("measurements", {})
            self.logs.debug(
                f"Anomaly Score: using historical measurements for identity {self.metrics[metric_name].identity}"
            )
            historic_check_results = historic_values.get(KEY_HISTORIC_CHECK_RESULTS, {}).get("check_results", {})
        else:
            self.logs.error(
                "Getting historical measurements and check results from Soda Cloud resulted in a "
                f"{type(historic_values)} object which is not compatible with anomaly detection. "
                "Check previous log messages for more information."
            )
            return

        if not historic_measurements:
            self.logs.warning(f"This is the first time that we derive {metrics[metric_name]} metric")
            historic_measurements["results"] = []

        # Append current results
        historic_measurements.get("results", []).append(
            {
                "id": str(61),  # Placeholder number that will be overwritten
                "identity": metrics[metric_name].identity,
                "value": self.get_metric_value(),
                "dataTime": (
                    self.data_source_scan.scan._data_timestamp.replace(tzinfo=timezone.utc).strftime(
                        "%Y-%m-%dT%H:%M:%SZ"
                    )
                ),
            }
        )

        # TODO test for module installation and set check status to is_skipped if the module is not installed
        try:
            from soda.scientific.anomaly_detection.anomaly_detector import (
                AnomalyDetector,
            )
        except ModuleNotFoundError as e:
            self.logs.error(f"{SODA_SCIENTIFIC_MISSING_LOG_MESSAGE}\n Original error: {e}")
            return

        warn_only = False
        if self.check_cfg.source_configurations is not None:
            warn_only = self.check_cfg.source_configurations.get("warn_only", False)

        anomaly_detector = AnomalyDetector(
            historic_measurements, historic_check_results, self.logs, metric_name, warn_only
        )
        level, diagnostics = anomaly_detector.evaluate()
        assert isinstance(diagnostics, dict), f"Anomaly diagnostics should be a dict. Got a {type(diagnostics)} instead"

        if diagnostics["anomalyErrorCode"] == "not_enough_measurements":
            self.add_outcome_reason(
                outcome_type=diagnostics["anomalyErrorCode"],
                message="Anomaly detection needs at least 5 measurements",
                severity=diagnostics["anomalyErrorSeverity"],
            )
            self.diagnostics = diagnostics
            if self.diagnostics["value"] is None:
                self.diagnostics["value"] = self.get_metric_value()
            return
        self.outcome = CheckOutcome(level)
        self.diagnostics = diagnostics
        if diagnostics["anomalyErrorSeverity"] in ["warn", "error"]:
            self.add_outcome_reason(
                outcome_type=diagnostics["anomalyErrorCode"],
                message=diagnostics["anomalyErrorMessage"],
                severity=diagnostics["anomalyErrorSeverity"],
            )

    def get_cloud_diagnostics_dict(self) -> dict:
        cloud_diagnostics = super().get_cloud_diagnostics_dict()
        return {**cloud_diagnostics, **self.diagnostics}

    def get_log_diagnostic_dict(self) -> dict:
        log_diagnostics = super().get_log_diagnostic_dict()
        if self.historic_diff_values:
            log_diagnostics.update(self.diagnostics)
        return log_diagnostics



from __future__ import annotations

from datetime import date, datetime, timedelta, timezone

from soda.execution.check.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.metric.metric import Metric
from soda.execution.metric.numeric_query_metric import NumericQueryMetric

MAX_COLUMN_TIMESTAMP = "max_column_timestamp"


class FreshnessCheck(Check):
    def __init__(
        self,
        check_cfg: FreshnessCheckCfg,
        data_source_scan: DataSourceScan,
        partition: Partition,
        column: Column,
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
        )
        self.freshness_values: dict | None = None
        self.metrics[MAX_COLUMN_TIMESTAMP] = data_source_scan.resolve_metric(
            NumericQueryMetric(
                data_source_scan=self.data_source_scan,
                partition=self.partition,
                column=self.column,
                metric_name="max",
                metric_args=None,
                filter=self.check_cfg.filter,
                aggregation=None,
                check_missing_and_valid_cfg=None,
                column_configurations_cfg=None,
                check=self,
            )
        )
        self.freshness_values = {}

    def evaluate(self, metrics: dict[str, Metric], historic_values: dict[str, object]):
        from soda.sodacl.freshness_check_cfg import FreshnessCheckCfg

        check_cfg: FreshnessCheckCfg = self.check_cfg

        now_variable_name = check_cfg.variable_name
        now_variable_timestamp_text = self.data_source_scan.scan.get_variable(now_variable_name)
        if not now_variable_timestamp_text:
            self.outcome = CheckOutcome.FAIL
            self.logs.error(
                f"Could not parse variable {now_variable_name} as a timestamp: variable not found",
            )
            return

        max_column_timestamp: datetime | None = metrics.get(MAX_COLUMN_TIMESTAMP).value
        now_variable_timestamp: datetime | None = None

        if type(max_column_timestamp) == date:  # using type instead of isinstance because datetime is subclass of date.
            # Convert data to datetime if its date, use max time (1ms before midnight)
            min_time = datetime.max.time()
            max_column_timestamp = datetime.combine(max_column_timestamp, min_time)

        is_max_column_timestamp_valid = isinstance(max_column_timestamp, datetime)
        if not is_max_column_timestamp_valid and max_column_timestamp is not None:
            self.logs.error(
                f"Could not evaluate freshness: max({self.check_cfg.column_name}) "
                f"is not a datetime: {type(max_column_timestamp).__name__}",
                location=self.check_cfg.location,
            )

        try:
            now_variable_timestamp = datetime.fromisoformat(now_variable_timestamp_text)
        except:
            self.logs.error(
                f"Could not parse variable {now_variable_name} as a timestamp: {now_variable_timestamp_text}",
                location=check_cfg.location,
            )

        is_now_variable_timestamp_valid = isinstance(now_variable_timestamp, datetime)

        freshness = None
        if not is_now_variable_timestamp_valid:
            self.logs.error(
                f"Could not evaluate freshness: variable {check_cfg.variable_name} is not a datetime: {now_variable_timestamp_text}",
                location=self.check_cfg.location,
            )

        max_column_timestamp_utc = (
            self._datetime_to_utc(max_column_timestamp) if isinstance(max_column_timestamp, datetime) else None
        )
        now_timestamp_utc = (
            self._datetime_to_utc(now_variable_timestamp) if isinstance(now_variable_timestamp, datetime) else None
        )

        if is_max_column_timestamp_valid and is_now_variable_timestamp_valid:
            freshness = now_timestamp_utc - max_column_timestamp_utc
            if check_cfg.fail_freshness_threshold is not None and freshness > check_cfg.fail_freshness_threshold:
                self.outcome = CheckOutcome.FAIL
            elif check_cfg.warn_freshness_threshold is not None and freshness > check_cfg.warn_freshness_threshold:
                self.outcome = CheckOutcome.WARN
            else:
                self.outcome = CheckOutcome.PASS
        elif max_column_timestamp is None:
            self.outcome = CheckOutcome.FAIL

        self.freshness_values = {
            "max_column_timestamp": str(max_column_timestamp) if max_column_timestamp else None,
            "max_column_timestamp_utc": str(max_column_timestamp_utc) if max_column_timestamp_utc else None,
            "now_variable_name": now_variable_name,
            "now_timestamp": now_variable_timestamp_text,
            "now_timestamp_utc": str(now_timestamp_utc) if now_timestamp_utc else None,
            "freshness": freshness,
        }

    def get_cloud_diagnostics_dict(self):
        cloud_diagnostics = super().get_cloud_diagnostics_dict()

        if self.freshness_values:
            freshness = 0
            if self.freshness_values["freshness"] and isinstance(self.freshness_values["freshness"], timedelta):
                freshness = round(self.freshness_values["freshness"].total_seconds() * 1000)

            cloud_diagnostics["value"] = freshness  # milliseconds difference
            cloud_diagnostics["measure"] = "time"
            cloud_diagnostics["maxColumnTimestamp"] = self.freshness_values["max_column_timestamp"]
            cloud_diagnostics["maxColumnTimestampUtc"] = self.freshness_values["max_column_timestamp_utc"]
            cloud_diagnostics["nowVariableName"] = self.freshness_values["now_variable_name"]
            cloud_diagnostics["nowTimestamp"] = self.freshness_values["now_timestamp"]
            cloud_diagnostics["nowTimestampUtc"] = self.freshness_values["now_timestamp_utc"]
            cloud_diagnostics["freshness"] = self.freshness_values["freshness"]

        return cloud_diagnostics

    def get_log_diagnostic_dict(self) -> dict:
        return self.freshness_values

    @staticmethod
    def _datetime_to_utc(input: datetime) -> datetime:
        if input.tzinfo is None:
            return input.replace(tzinfo=timezone.utc)
        return input.astimezone(timezone.utc)



from soda.execution.query.query import Query


class GroupByQuery(Query):
    def __init__(
        self,
        data_source_scan: "DataSourceScan",
        metric: "GroupByMetric",
        partition: "Partition",
        location: "Location",
        group_limit: int = 1000,
    ):
        super().__init__(
            data_source_scan=data_source_scan,
            unqualified_query_name=f"group_by[{metric.name}]",
            location=location,
            sql=metric.query,
            partition=partition,
        )
        self.group_limit = group_limit
        self.metric = metric

    def execute(self):
        self.fetchall()

        results = []
        columns = [column[0] for column in self.description]
        for row in self.rows:
            results.append(dict(zip(columns, row)))
        self.metric.set_value(results)



from abc import ABC, abstractmethod
from typing import Tuple

from soda.sampler.sample_schema import SampleSchema


class Sample(ABC):
    @abstractmethod
    def get_rows(self) -> Tuple[Tuple]:
        # get_rows should return a sufficient number of rows to fulfill
        # the sample request, but is not guaranteed to return all rows.
        # Leverage
        pass

    @abstractmethod
    def get_rows_count(self) -> int:
        # Returns total number of rows involved with the sampler. This number
        # can be higher than the total number of rows returned by get_rows.
        pass

    @abstractmethod
    def get_schema(self) -> SampleSchema:
        pass



from __future__ import annotations

from soda.execution.query.query import Query


class UserDefinedFailedRowsExpressionQuery(Query):
    def __init__(
        self,
        data_source_scan: DataSourceScan,
        check_name: str,
        sql: str,
        partition: Partition,
        samples_limit: int | None = None,
        metric: Metric = None,
    ):
        super().__init__(
            data_source_scan=data_source_scan,
            sql=sql,
            unqualified_query_name=f"user_defined_failed_rows_expression_query[{check_name}]",
            partition=partition,
        )
        self.samples_limit = samples_limit
        self.metric = metric

    def execute(self):
        self.store()



from __future__ import annotations

import datetime
import hashlib
import importlib
import json
import re
from collections import defaultdict
from datetime import date, datetime
from functools import lru_cache
from numbers import Number
from textwrap import dedent

from soda.common.exceptions import DataSourceError
from soda.common.logs import Logs
from soda.common.string_helper import string_matches_simple_pattern
from soda.execution.data_type import DataType
from soda.execution.query.query import Query
from soda.execution.query.query_without_results import QueryWithoutResults
from soda.execution.query.schema_query import TableColumnsQuery
from soda.sampler.sample_ref import SampleRef
from soda.sodacl.location import Location


class FormatHelper:
    @staticmethod
    def build_default_formats():
        # s stands for an optional set of spaces
        s = r" *"
        sign_any_opt = r"[-+]?"
        plus_opt = r"\+?"
        min = r"\-"
        digits = r"[0-9]+"
        decimal_point_comma = r"(\d+([\.,]\d+)?|([\.,]\d+))"
        decimal_point = r"(\d+(\.\d+)?|(\.\d+))"
        decimal_comma = r"(\d+(,\d+)?|(,\d+))"
        zero_integer = r"0+"
        zero_point_comma = r"(0+([\.,]0+)?|([\.,]0+))"
        zero_point = r"(0+(\.0+)?|(\.0+))"
        zero_comma = r"(0+(,0+)?|(,0+))"
        money_point = r"\d{1,3}(\,\d\d\d)*(.\d+)?"
        money_comma = r"\d{1,3}(\.\d\d\d)*(,\d+)?"
        currency = r"([A-Z]{3}|[a-z]{3})"

        day = r"([1-9]|[012][0-9]|3[01])"
        month = r"([1-9]|0[1-9]|1[012])"
        year = r"(19|20)?\d\d"
        hour24 = r"(0?[0-9]|[01]\d|2[0-3])"
        hour12 = r"(0?[0-9]|1[0-2])"
        minute = r"[0-5]?[0-9]"
        second = r"[0-5]?[0-9]([.,]\d+)?"
        year4 = r"(19|20)\d\d"
        month2 = r"(0[0-9]|1[12])"
        day2 = r"([012][0-9]|3[01])"
        hour2 = r"(0[0-9]|1[012])"
        minute2 = r"[0-5][0-9]"
        second2 = minute2

        return {
            "integer": f"^{s}{sign_any_opt}{s}{digits}{s}$",
            "positive integer": f"^{s}{plus_opt}{s}{digits}{s}$",
            "negative integer": f"^{s}({min}{s}{digits}|{zero_integer}){s}$",
            "decimal": f"^{s}{sign_any_opt}{s}{decimal_point_comma}{s}$",
            "positive decimal": f"^{s}{plus_opt}{s}{decimal_point_comma}{s}$",
            "negative decimal": f"^{s}({min}{s}{decimal_point_comma}|{zero_point_comma}){s}$",
            "decimal point": f"^{s}{sign_any_opt}{s}{decimal_point}{s}$",
            "positive decimal point": f"^{s}{plus_opt}{s}{decimal_point}{s}$",
            "negative decimal point": f"^{s}({min}{s}{decimal_point}|{zero_point}){s}$",
            "decimal comma": f"^{s}{sign_any_opt}{s}{decimal_comma}{s}$",
            "positive decimal comma": f"^{s}{plus_opt}{s}{decimal_comma}{s}$",
            "negative decimal comma": f"^{s}({min}{s}{decimal_comma}|{zero_comma}){s}$",
            "percentage": f"^{s}{sign_any_opt}{s}{decimal_point_comma}{s}%{s}$",
            "positive percentage": f"^{s}{plus_opt}{s}{decimal_point_comma}{s}%{s}$",
            "negative percentage": f"^{s}({min}{s}{decimal_point_comma}|{zero_point_comma}){s}%{s}$",
            "percentage point": f"^{s}{sign_any_opt}{s}{decimal_point}{s}%{s}$",
            "positive percentage point": f"^{s}{plus_opt}{s}{decimal_point}{s}%{s}$",
            "negative percentage point": f"^{s}({min}{s}{decimal_point}|{zero_point}){s}%{s}$",
            "percentage comma": f"^{s}{sign_any_opt}{s}{decimal_comma}{s}%{s}$",
            "positive percentage comma": f"^{s}{plus_opt}{s}{decimal_comma}{s}%{s}$",
            "negative percentage comma": f"^{s}({min}{s}{decimal_comma}|{zero_comma}){s}%{s}$",
            "money": f"^{s}{sign_any_opt}{s}{decimal_point_comma}{s}{currency}{s}$",
            "money point": f"^{s}{sign_any_opt}{s}{money_point}{s}{currency}{s}$",
            "money comma": f"^{s}{sign_any_opt}{s}{money_comma}{s}{currency}{s}$",
            "date us": rf"^{s}{month}[-\./]{day}[-\./]{year}{s}$",
            "date eu": rf"^{s}{day}[-\./]{month}[-\./]{year}{s}$",
            "date inverse": rf"^{s}{year}[-\./]{month}[-\./]{day}{s}$",
            "date iso 8601": f"^{s}"
            rf"{year4}-?({month2}-?{day2}|W[0-5]\d(-?[1-7])?|[0-3]\d\d)"
            rf"([ T]{hour2}(:?{minute2}(:?{second2}([.,]\d+)?)?)?([+-]{hour2}:?{minute2}|Z)?)?"
            f"{s}$",
            "time 24h": f"^{s}{hour24}:{minute}(:{second})?{s}$",
            "time 24h nosec": f"^{s}{hour24}:{minute}{s}$",
            "time 12h": f"^{s}{hour12}:{minute}(:{second})?{s}$",
            "time 12h nosec": f"^{s}{hour12}:{minute}{s}$",
            "timestamp 24h": f"^{s}{hour24}:{minute}:{second}{s}$",
            "timestamp 12h": f"^{s}{hour12}:{minute}:{second}{s}$",
            "uuid": r"^[0-9a-fA-F]{8}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{12}$",
            "ip address": r"^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$",
            "ipv4 address": r"^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$",
            # scary - but covers all cases
            "ipv6 address": r"^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))$",
            "email": r"^[a-zA-ZÀ-ÿĀ-ſƀ-ȳ0-9.\-_%+]+@[a-zA-ZÀ-ÿĀ-ſƀ-ȳ0-9.\-_%]+\.[A-Za-z]{2,4}$",
            "phone number": r"^((\+[0-9]{1,2}\s)?\(?[0-9]{3}\)?[\s.-])?[0-9]{3}[\s.-][0-9]{4}$",
            "credit card number": r"^[0-9]{14}|[0-9]{15}|[0-9]{16}|[0-9]{17}|[0-9]{18}|[0-9]{19}|([0-9]{4}-){3}[0-9]{4}|([0-9]{4} ){3}[0-9]{4}$",
        }


class DataSource:
    """
    Data source implementation.

    For documentation on implementing a new DataSource see the CONTRIBUTING-DATA-SOURCE.md document.
    """

    # Maps synonym types for the convenience of use in checks.
    # Keys represent the data_source type, values are lists of "aliases" that can be used in SodaCL as synonyms.
    SCHEMA_CHECK_TYPES_MAPPING: dict = {
        "character varying": ["varchar", "text"],
        "double precision": ["decimal", "numeric"],
        "timestamp without time zone": ["timestamp"],
        "timestamp with time zone": ["timestamptz"],
    }

    # Supported data types used in create statements. These are used in tests for creating test data and do not affect the actual library functionality.
    SQL_TYPE_FOR_CREATE_TABLE_MAP: dict = {
        DataType.TEXT: "VARCHAR(255)",
        DataType.INTEGER: "INT",
        DataType.DECIMAL: "FLOAT",
        DataType.DATE: "DATE",
        DataType.TIME: "TIME",
        DataType.TIMESTAMP: "TIMESTAMP",
        DataType.TIMESTAMP_TZ: "TIMESTAMPTZ",
        DataType.BOOLEAN: "BOOLEAN",
    }

    # Supported data types as returned by the given data source when retrieving dataset schema. Used in schema checks.
    SQL_TYPE_FOR_SCHEMA_CHECK_MAP = {
        DataType.TEXT: "character varying",
        DataType.INTEGER: "integer",
        DataType.DECIMAL: "double precision",
        DataType.DATE: "date",
        DataType.TIME: "time",
        DataType.TIMESTAMP: "timestamp without time zone",
        DataType.TIMESTAMP_TZ: "timestamp with time zone",
        DataType.BOOLEAN: "boolean",
    }

    # Indicate which numeric/test data types can be used for profiling checks.
    NUMERIC_TYPES_FOR_PROFILING = [
        "integer",
        "double precision",
        "double",
        "smallint",
        "bigint",
        "decimal",
        "numeric",
        "real",
        "smallserial",
        "serial",
        "bigserial",
    ]
    TEXT_TYPES_FOR_PROFILING = ["character varying", "varchar", "text", "character", "char"]
    LIMIT_KEYWORD: str = "LIMIT"

    # Building up format queries normally works with regexp expression + a set of formats,
    # but some use cases require whole completely custom format expressions.
    # DEFAULT_FORMAT_EXPRESSIONS take precedence over DEFAULT_FORMATS.
    DEFAULT_FORMATS: dict[str, str] = FormatHelper.build_default_formats()
    DEFAULT_FORMAT_EXPRESSIONS: dict[str, str] = {}

    @staticmethod
    def camel_case_data_source_type(data_source_type: str) -> str:
        if "bigquery" == data_source_type:
            return "BigQuery"
        elif "spark_df" == data_source_type:
            return "SparkDf"
        elif "sqlserver" == data_source_type:
            return "SQLServer"
        elif "mysql" == data_source_type:
            return "MySQL"
        elif "duckdb" == data_source_type:
            return "DuckDB"
        else:
            return f"{data_source_type[0:1].upper()}{data_source_type[1:]}"

    @staticmethod
    def create(
        logs: Logs,
        data_source_name: str,
        data_source_type: str,
        data_source_properties: dict,
    ) -> DataSource:
        """
        The returned data_source does not have a connection.  It is the responsibility of
        the caller to initialize data_source.connection.  To create a new connection,
        use data_source.connect(...)
        """
        try:
            data_source_properties["connection_type"] = data_source_type
            module = importlib.import_module(f"soda.data_sources.{data_source_type}_data_source")
            data_source_class = f"{DataSource.camel_case_data_source_type(data_source_type)}DataSource"
            class_ = getattr(module, data_source_class)
            return class_(logs, data_source_name, data_source_properties)
        except ModuleNotFoundError:
            if data_source_type == "postgresql":
                logs.error(f'Data source type "{data_source_type}" not found. Did you mean postgres?')
            else:
                raise DataSourceError(
                    f'Data source type "{data_source_type}" not found. Did you spell {data_source_type} correctly? Did you install module soda-core-{data_source_type}?'
                )
            return None

    def __init__(
        self,
        logs: Logs,
        data_source_name: str,
        data_source_properties: dict,
    ):
        self.host = data_source_properties.get("host")
        self.logs = logs
        self.data_source_name = data_source_name
        self.data_source_properties: dict = data_source_properties
        # Pep 249 compliant connection object (aka DBAPI)
        # https://www.python.org/dev/peps/pep-0249/#connection-objects
        # @see self.connect() for initialization
        self.type = self.data_source_properties.get("connection_type")
        self.connection = None
        self.database: str | None = data_source_properties.get("database")
        self.schema: str | None = data_source_properties.get("schema")
        self.table_prefix: str | None = self._create_table_prefix()
        # self.data_source_scan is initialized in create_data_source_scan(...) below
        self.data_source_scan: DataSourceScan | None = None
        # Temporarily introduced to migrate some "wrongly implemented" data sources.
        # See https://sodadata.atlassian.net/browse/CLOUD-5446
        self.migrate_data_source_name = None
        self.quote_tables: bool = data_source_properties.get("quote_tables", False)

    def has_valid_connection(self) -> bool:
        query = Query(
            data_source_scan=self.data_source_scan,
            sql=self.sql_test_connection(),
            unqualified_query_name="test-connection",
        )
        query.execute()

        if query.exception:
            return False

        return True

    def create_data_source_scan(self, scan: Scan, data_source_scan_cfg: DataSourceScanCfg):
        from soda.execution.data_source_scan import DataSourceScan

        data_source_scan = DataSourceScan(scan, data_source_scan_cfg, self)
        self.data_source_scan = data_source_scan

        return self.data_source_scan

    def validate_configuration(self, logs: Logs) -> None:
        raise NotImplementedError(f"TODO: Implement {type(self)}.validate_configuration(...)")

    def get_type_name(self, type_code):
        return str(type_code)

    def is_supported_metric_name(self, metric_name: str) -> bool:
        return (
            metric_name in ["row_count", "missing_count", "invalid_count", "valid_count", "duplicate_count"]
            or self.get_metric_sql_aggregation_expression(metric_name, None, None) is not None
        )

    def get_metric_sql_aggregation_expression(self, metric_name: str, metric_args: list[object] | None, expr: str):
        if "min" == metric_name:
            return self.expr_min(expr)
        if "max" == metric_name:
            return self.expr_max(expr)
        if "avg" == metric_name:
            return self.expr_avg(expr)
        if "sum" == metric_name:
            return self.expr_sum(expr)
        if "min_length" == metric_name:
            return self.expr_min(self.expr_length(expr))
        if "max_length" == metric_name:
            return self.expr_max(self.expr_length(expr))
        if "avg_length" == metric_name:
            return self.expr_avg(self.expr_length(expr))
        return None

    def is_same_type_in_schema_check(self, expected_type: str, actual_type: str):
        expected_type = expected_type.strip().lower()

        if (
            actual_type in self.SCHEMA_CHECK_TYPES_MAPPING
            and expected_type in self.SCHEMA_CHECK_TYPES_MAPPING[actual_type]
        ):
            return True

        if (
            expected_type in self.SCHEMA_CHECK_TYPES_MAPPING
            and actual_type in self.SCHEMA_CHECK_TYPES_MAPPING[expected_type]
        ):
            return True

        return expected_type == actual_type.lower()

    @staticmethod
    def column_metadata_columns() -> list:
        """Columns to be used for retrieving column metadata."""
        return ["column_name", "data_type", "is_nullable"]

    @staticmethod
    def column_metadata_catalog_column() -> str:
        """Column to be used as a 'database' equivalent."""
        return "table_catalog"

    @staticmethod
    def column_metadata_table_name() -> str:
        return "table_name"

    @staticmethod
    def column_metadata_schema_name() -> str:
        return "table_schema"

    @staticmethod
    def column_metadata_column_name() -> str:
        return "column_name"

    @staticmethod
    def column_metadata_datatype_name() -> str:
        return "data_type"

    def tables_columns_metadata(self) -> list[str]:
        """Columns to be used for retrieving tables and columns metadata."""
        return [
            self.column_metadata_table_name(),
            self.column_metadata_column_name(),
            self.column_metadata_datatype_name(),
        ]

    ######################
    # Store Table Sample
    ######################

    def store_table_sample(self, table_name: str, limit: int | None = None, filter: str | None = None) -> SampleRef:
        sql = self.sql_select_all(table_name=table_name, limit=limit, filter=filter)
        query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name=f"store-sample-for-{table_name}",
            sql=sql,
            sample_name="table_sample",
        )
        query.store()
        return query.sample_ref

    def sql_select_all(self, table_name: str, limit: int | None = None, filter: str | None = None) -> str:
        qualified_table_name = self.qualified_table_name(table_name)

        filter_sql = ""
        if filter:
            filter_sql = f" \n WHERE {filter}"

        limit_sql = ""
        if limit is not None:
            limit_sql = f" \n LIMIT {limit}"

        columns_names = ", ".join(self.sql_select_all_column_names(table_name))

        sql = f"SELECT {columns_names} FROM {qualified_table_name}{filter_sql}{limit_sql}"
        return sql

    def sql_select_all_column_names(self, table_name: str) -> list:
        selectable_columns = []

        if self.get_exclude_column_patterns_for_table(table_name) and self.data_source_scan.scan._configuration.sampler:
            all_columns = self.get_table_columns(table_name, f"get_table_columns_{table_name}")
            exclude_columns = []

            for column in all_columns:
                if self.is_column_excluded(table_name, column):
                    exclude_columns.append(column)
                else:
                    selectable_columns.append(column)

            if exclude_columns:
                self.logs.debug(
                    f"Skipping columns {exclude_columns} from table '{table_name}' when selecting all columns data."
                )
            if not selectable_columns:
                self.logs.info(
                    f"Unable to select data for failed rows from table '{table_name}', all columns are excluded. Selecting '*' for the check, no failed rows samples will be created."
                )

        if not selectable_columns:
            selectable_columns = ["*"]

        return selectable_columns

    def get_exclude_column_patterns_for_table(self, table_name: str) -> list[str]:
        """Match table and column names case insensitive."""
        exclude_columns_config = {
            k.lower(): [c.lower() for c in v]
            for k, v in self.data_source_scan.scan._configuration.exclude_columns.items()
        }
        exclude_column_patterns = []

        table_name_lower = table_name.lower()

        for table_pattern, column_patterns in exclude_columns_config.items():
            "Table sometimes comes with a schema a prefix, try matching on 'prefix.pattern' as well"
            table_pattern_with_schema = f"{self.table_prefix if self.table_prefix else ''}.{table_pattern}".lower()
            if string_matches_simple_pattern(table_name_lower, table_pattern) or string_matches_simple_pattern(
                table_name_lower, table_pattern_with_schema
            ):
                exclude_column_patterns += column_patterns

        return list(set(exclude_column_patterns))

    def is_column_excluded(self, table_name: str, column: str) -> bool:
        column_patterns_for_table = self.get_exclude_column_patterns_for_table(table_name)

        for pattern in column_patterns_for_table:
            if string_matches_simple_pattern(column, pattern) or column == "*":
                # '*' is a special case - it is considered excluded in case there is at least one column exclude pattern for the given table.
                return True

        return False

    @staticmethod
    def parse_tables_columns_query(rows: list[tuple]) -> defaultdict(dict):
        tables_and_columns_metadata = defaultdict(dict)
        for table_name, column_name, data_type in rows:
            tables_and_columns_metadata[table_name][column_name] = data_type
        return tables_and_columns_metadata

    def get_tables_columns_metadata(
        self,
        query_name: str,
        include_patterns: list[dict[str, str]] | None = None,
        exclude_patterns: list[dict[str, str]] | None = None,
        table_names_only: bool = False,
    ) -> defaultdict[str, dict[str, str]] | None:
        # TODO: save/cache the result for later use.
        if (not include_patterns) and (not exclude_patterns):
            return []
        query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name=query_name,
            sql=self.sql_get_tables_columns_metadata(
                include_patterns=include_patterns, exclude_patterns=exclude_patterns, table_names_only=table_names_only
            ),
        )
        query.execute()

        rows = query.rows
        if rows and len(rows) > 0:
            if table_names_only:
                query_result = [self._optionally_quote_table_name_from_meta_data(row[0]) for row in rows]
            else:
                query_result: defaultdict[dict] = self.parse_tables_columns_query(rows)
            return query_result
        return None

    def create_table_column_sql_filters(
        self, sql_patterns: list[dict[str, str]], table_names_only: bool = False
    ) -> list[str]:
        sql_filters = []

        for pattern in sql_patterns:
            sql_filter = []

            table_name_pattern = pattern.get("table_name_pattern")
            if table_name_pattern is not None:
                table_name_filter = (
                    f"({self.column_metadata_table_name()} LIKE '{self.default_casify_table_name(table_name_pattern)}')"
                )
                sql_filter.append(table_name_filter)

            column_name_pattern = pattern.get("column_name_pattern")
            if (column_name_pattern is not None) and (table_names_only is False):
                column_name_filter = f"({self.default_casify_sql_function()}({self.column_metadata_column_name()}) LIKE {self.default_casify_sql_function()}('{column_name_pattern}'))"
                sql_filter.append(column_name_filter)

            if sql_filter:
                sql_filters.append(f"""({" AND ".join(sql_filter)})""")

        return sql_filters

    def catalog_column_filter(self):
        catalog_filter = f"{self.default_casify_sql_function()}({self.column_metadata_catalog_column()}) = '{self.default_casify_system_name(self.database)}'"
        return catalog_filter

    def sql_get_tables_columns_metadata(
        self,
        include_patterns: list[dict[str, str]] | None = None,
        exclude_patterns: list[dict[str, str]] | None = None,
        table_names_only: bool = False,
    ) -> str:
        filter_clauses = []

        if include_patterns and len(include_patterns) > 0:
            include_sql_filter_clauses = self.create_table_column_sql_filters(
                include_patterns, table_names_only=table_names_only
            )
            include_filter = " OR ".join(include_sql_filter_clauses)
            filter_clauses.append(f"({include_filter})")

        if exclude_patterns and len(exclude_patterns) > 0:
            exclude_sql_filter_clauses = [
                f"NOT {sql_filter_clause}"
                for sql_filter_clause in self.create_table_column_sql_filters(
                    exclude_patterns, table_names_only=table_names_only
                )
            ]
            exclude_filter = " AND ".join(exclude_sql_filter_clauses)
            filter_clauses.append(f"({exclude_filter})")

        if self.database:
            catalog_filter = self.catalog_column_filter()
            if catalog_filter:
                filter_clauses.append(catalog_filter)

        if hasattr(self, "schema") and self.schema:
            filter_clauses.append(
                f"{self.default_casify_sql_function()}({self.column_metadata_schema_name()}) = '{self.default_casify_system_name(self.schema)}'"
            )

        where_filter = " \n  AND ".join(filter_clauses)

        # compose query template
        # NOTE: we use `order by ordinal_position` to guarantee stable orders of columns
        # (see https://www.postgresql.org/docs/current/infoschema-columns.html)
        # this mainly has an advantage in testing but bears very little as to how Soda Cloud
        # displays those columns as they are ordered alphabetically in the UI.

        if table_names_only:
            metadata_columns = f"{self.column_metadata_table_name()}"
            information_schema_table = self.sql_information_schema_tables()
            order_by_clause = ""
        else:
            metadata_columns = ", ".join(self.tables_columns_metadata())
            information_schema_table = self.sql_information_schema_columns()
            order_by_clause = f"\nORDER BY {self.get_ordinal_position_name()}"

        sql = (
            f"SELECT {metadata_columns} \n"
            f"FROM {information_schema_table} \n"
            f"WHERE {where_filter}"
            f"{order_by_clause}"
        )

        return sql

    ############################################
    # For a table, get the columns metadata
    ############################################

    @lru_cache(maxsize=None)
    def get_table_columns(
        self,
        table_name: str,
        query_name: str,
        included_columns: list[str] | None = None,
        excluded_columns: list[str] | None = None,
    ) -> dict[str, str]:
        """
        :return: A dict mapping column names to data source data types.  Like eg
        {"id": "varchar", "cst_size": "int8", ...}
        """
        query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name=query_name,
            sql=self.sql_get_table_columns(
                table_name, included_columns=included_columns, excluded_columns=excluded_columns
            ),
        )
        query.execute()
        if query.rows and len(query.rows) > 0:
            return {row[0]: row[1] for row in query.rows}
        return {}

    def create_table_columns_query(self, partition: Partition, schema_metric: SchemaMetric) -> TableColumnsQuery:
        return TableColumnsQuery(partition, schema_metric)

    def get_ordinal_position_name(self) -> str:
        return "ORDINAL_POSITION"

    def sql_get_table_columns(
        self,
        table_name: str,
        included_columns: list[str] | None = None,
        excluded_columns: list[str] | None = None,
    ) -> str:
        table_name_default_case = self.default_casify_table_name(table_name)
        unquoted_table_name_default_case = (
            table_name_default_case[1:-1] if self.is_quoted(table_name_default_case) else table_name_default_case
        )

        casify_function = self.default_casify_sql_function()
        filter_clauses = [f"{casify_function}(table_name) = '{unquoted_table_name_default_case}'"]

        if self.database:
            filter_clauses.append(
                f"{casify_function}({self.column_metadata_catalog_column()}) = '{self.default_casify_system_name(self.database)}'"
            )

        if self.schema:
            filter_clauses.append(
                f"{casify_function}({self.column_metadata_schema_name()}) = '{self.default_casify_system_name(self.schema)}'"
            )

        if included_columns:
            include_clauses = []
            for col in included_columns:
                include_clauses.append(f"{casify_function}(column_name) LIKE {casify_function}('{col}')")
            include_causes_or = " OR ".join(include_clauses)
            filter_clauses.append(f"({include_causes_or})")

        if excluded_columns:
            for col in excluded_columns:
                filter_clauses.append(f"{casify_function}(column_name) NOT LIKE {casify_function}('{col}')")

        where_filter = " \n  AND ".join(filter_clauses)

        # compose query template
        # NOTE: we use `order by ordinal_position` to guarantee stable orders of columns
        # (see https://www.postgresql.org/docs/current/infoschema-columns.html)
        # this mainly has an advantage in testing but bears very little as to how Soda Cloud
        # displays those columns as they are ordered alphabetically in the UI.
        sql = (
            f"SELECT {', '.join(self.column_metadata_columns())} \n"
            f"FROM {self.sql_information_schema_columns()} \n"
            f"WHERE {where_filter}"
            f"\nORDER BY {self.get_ordinal_position_name()}"
        )
        return sql

    ############################################
    # Get table names with count in one go
    ############################################

    def sql_get_table_names_with_count(
        self, include_tables: list[str] | None = None, exclude_tables: list[str] | None = None
    ) -> str:
        table_filter_expression = self.sql_table_include_exclude_filter(
            "relname", "schemaname", include_tables, exclude_tables
        )
        where_clause = f"\nWHERE {table_filter_expression} \n" if table_filter_expression else ""
        return f"SELECT relname, n_live_tup \n" f"FROM pg_stat_user_tables" f"{where_clause}"

    def sql_get_table_count(self, table_name: str) -> str:
        return f"SELECT {self.expr_count_all()} from {self.qualified_table_name(table_name)}"

    def sql_table_include_exclude_filter(
        self,
        table_column_name: str,
        schema_column_name: str | None = None,
        include_tables: list[str] | None = [],
        exclude_tables: list[str] | None = [],
    ) -> str | None:
        tablename_filter_clauses = []

        def build_table_matching_conditions(tables: list[str], comparison_operator: str):
            conditions = []

            for table in tables:
                # This is intended to be future proof and support quoted table names. The method is not used in such way and table names/patterns that still
                # need to be quoted are passed here, e.g. `%sodatest_%`. I.e. this condition is always met and default casify is always used, table name below is therefore single quoted.
                if not self.is_quoted(table):
                    table = self.default_casify_table_name(table)

                conditions.append(f"{table_column_name} {comparison_operator} '{table}'")
            return conditions

        if include_tables:
            sql_include_clauses = " OR ".join(build_table_matching_conditions(include_tables, "like"))
            tablename_filter_clauses.append(f"({sql_include_clauses})")

        if exclude_tables:
            tablename_filter_clauses.extend(build_table_matching_conditions(exclude_tables, "not like"))

        if hasattr(self, "schema") and self.schema and schema_column_name:
            tablename_filter_clauses.append(f"lower({schema_column_name}) = '{self.schema.lower()}'")

        return "\n      AND ".join(tablename_filter_clauses) if tablename_filter_clauses else None

    def sql_find_table_names(
        self,
        filter: str | None = None,
        include_tables: list[str] = [],
        exclude_tables: list[str] = [],
        table_column_name: str = "table_name",
        schema_column_name: str = "table_schema",
    ) -> str:
        sql = f"SELECT {table_column_name} \n" f"FROM {self.sql_information_schema_tables()}"
        where_clauses = []

        if filter:
            where_clauses.append(f"lower({self.default_casify_column_name(table_column_name)}) like '{filter.lower()}'")

        includes_excludes_filter = self.sql_table_include_exclude_filter(
            table_column_name, schema_column_name, include_tables, exclude_tables
        )
        if includes_excludes_filter:
            where_clauses.append(includes_excludes_filter)

        if where_clauses:
            where_clauses_sql = "\n  AND ".join(where_clauses)
            sql += f"\nWHERE {where_clauses_sql}"

        return sql

    def sql_information_schema_tables(self) -> str:
        return "information_schema.tables"

    def sql_information_schema_columns(self) -> str:
        return "information_schema.columns"

    def sql_analyze_table(self, table: str) -> str | None:
        return None

    def sql_get_duplicates_count(
        self,
        column_names: str,
        table_name: str,
        filter: str,
    ) -> str | None:
        qualified_table_name = self.qualified_table_name(table_name)

        sql = dedent(
            f"""
            WITH frequencies AS (
                SELECT {self.expr_count_all()} AS frequency
                FROM {qualified_table_name}
                WHERE {filter}
                GROUP BY {column_names})
            SELECT {self.expr_count_all()}
            FROM frequencies
            WHERE frequency > 1"""
        )

        return sql

    def sql_get_duplicates_aggregated(
        self,
        column_names: str,
        table_name: str,
        filter: str,
        limit: str | None = None,
        invert_condition: bool = False,
        exclude_patterns: list[str] | None = None,
    ) -> str | None:
        qualified_table_name = self.qualified_table_name(table_name)
        main_query_columns = f"{column_names}, frequency" if exclude_patterns else "*"

        sql = dedent(
            f"""
            WITH frequencies AS (
                SELECT {column_names}, {self.expr_count_all()} AS frequency
                FROM {qualified_table_name}
                WHERE {filter}
                GROUP BY {column_names})
            SELECT {main_query_columns}
            FROM frequencies
            WHERE frequency {'<=' if invert_condition else '>'} 1
            ORDER BY frequency DESC"""
        )

        if limit:
            sql += f"\nLIMIT {limit}"

        return sql

    def sql_get_duplicates(
        self,
        column_names: str,
        table_name: str,
        filter: str,
        limit: str | None = None,
        invert_condition: bool = False,
    ) -> str | None:
        qualified_table_name = self.qualified_table_name(table_name)
        columns = column_names.split(", ")

        main_query_columns = self.sql_select_all_column_names(table_name)
        qualified_main_query_columns = ", ".join([f"main.{c}" for c in main_query_columns])
        join = " AND ".join([f"main.{c} = frequencies.{c}" for c in columns])

        sql = dedent(
            f"""
            WITH frequencies AS (
                SELECT {column_names}
                FROM {qualified_table_name}
                WHERE {filter}
                GROUP BY {column_names}
                HAVING {self.expr_count_all()} {'<=' if invert_condition else '>'} 1)
            SELECT {qualified_main_query_columns}
            FROM {qualified_table_name} main
            JOIN frequencies ON {join}
            """
        )

        if limit:
            sql += f"\nLIMIT {limit}"

        return sql

    def sql_reference_query(
        self,
        columns: str,
        table_name: str,
        target_table_name: str,
        join_condition: str,
        where_condition: str,
        limit: int | None = None,
    ) -> str:
        sql = dedent(
            f"""
            SELECT {columns}
                FROM {table_name}  SOURCE
                LEFT JOIN {target_table_name} TARGET on {join_condition}
            WHERE {where_condition}"""
        )

        if limit:
            sql += f"\nLIMIT {limit}"

        return sql

    def cast_to_text(self, expr: str) -> str:
        return f"CAST({expr} AS VARCHAR)"

    def profiling_sql_values_frequencies_query(
        self,
        data_type_category: str,
        table_name: str,
        column_name: str,
        limit_mins_maxs: int,
        limit_frequent_values: int,
    ) -> str:
        cast_to_text = self.cast_to_text

        value_frequencies_cte = self.profiling_sql_value_frequencies_cte(table_name, column_name)

        union = self.sql_union()

        frequent_values_cte = f"""frequent_values AS (
                            SELECT {cast_to_text("'frequent_values'")} AS metric_, ROW_NUMBER() OVER(ORDER BY frequency_ DESC) AS index_, value_, frequency_
                            FROM value_frequencies
                            ORDER BY frequency_ desc
                            LIMIT {limit_frequent_values}
                        )"""

        if data_type_category == "text":
            return dedent(
                f"""
                    WITH
                        {value_frequencies_cte},
                        {frequent_values_cte}
                    SELECT *
                    FROM frequent_values
                    ORDER BY metric_ ASC, index_ ASC
                """
            )

        elif data_type_category == "numeric":
            mins_cte = f"""mins AS (
                            SELECT {cast_to_text("'mins'")} AS metric_, ROW_NUMBER() OVER(ORDER BY value_ ASC) AS index_, value_, frequency_
                            FROM value_frequencies
                            WHERE value_ IS NOT NULL
                            ORDER BY value_ ASC
                            LIMIT {limit_mins_maxs}
                        )"""

            maxs_cte = f"""maxs AS (
                            SELECT {cast_to_text("'maxs'")} AS metric_, ROW_NUMBER() OVER(ORDER BY value_ DESC) AS index_, value_, frequency_
                            FROM value_frequencies
                            WHERE value_ IS NOT NULL
                            ORDER BY value_ DESC
                            LIMIT {limit_mins_maxs}
                        )"""

            return dedent(
                f"""
                    WITH
                        {value_frequencies_cte},
                        {mins_cte},
                        {maxs_cte},
                        {frequent_values_cte},
                        results AS (
                            SELECT * FROM mins
                            {union}
                            SELECT * FROM maxs
                            {union}
                            SELECT * FROM frequent_values
                        )
                    SELECT *
                    FROM results
                    ORDER BY metric_ ASC, index_ ASC
                """
            )

        raise AssertionError("data_type_category must be either 'numeric' or 'text'")

    def sql_union(self):
        return "UNION"

    def profiling_sql_value_frequencies_cte(self, table_name: str, column_name: str) -> str:
        quoted_column_name = self.quote_column(column_name)
        qualified_table_name = self.qualified_table_name(table_name)
        return f"""value_frequencies AS (
                            SELECT {quoted_column_name} AS value_, {self.expr_count_all()} AS frequency_
                            FROM {qualified_table_name}
                            WHERE {quoted_column_name} IS NOT NULL
                            GROUP BY {quoted_column_name}
                        )"""

    def profiling_sql_aggregates_numeric(self, table_name: str, column_name: str) -> str:
        column_name = self.quote_column(column_name)
        qualified_table_name = self.qualified_table_name(table_name)
        return dedent(
            f"""
            SELECT
                avg({column_name}) as average
                , sum({column_name}) as sum
                , var_samp({column_name}) as variance
                , stddev_samp({column_name}) as standard_deviation
                , {self.expr_count(f'distinct({column_name})')} as distinct_values
                , sum(case when {column_name} is null then 1 else 0 end) as missing_values
            FROM {qualified_table_name}
            """
        )

    def profiling_sql_aggregates_text(self, table_name: str, column_name: str) -> str:
        column_name = self.quote_column(column_name)
        qualified_table_name = self.qualified_table_name(table_name)
        return dedent(
            f"""
            SELECT
                {self.expr_count(f'distinct({column_name})')} as distinct_values
                , sum(case when {column_name} is null then 1 else 0 end) as missing_values
                , avg(length({column_name})) as avg_length
                , min(length({column_name})) as min_length
                , max(length({column_name})) as max_length
            FROM {qualified_table_name}
            """
        )

    def histogram_sql_and_boundaries(
        self,
        table_name: str,
        column_name: str,
        min_value: int | float,
        max_value: int | float,
        n_distinct: int,
        column_type: str,
    ) -> tuple[str | None, list[int | float]]:
        # TODO: make configurable or derive dynamically based on data quantiles etc.
        max_n_bins = 20
        number_of_bins: int = max(1, min(n_distinct, max_n_bins))
        number_of_intervals: int = number_of_bins - 1

        if min_value >= max_value:
            self.logs.warning(
                "Min of {column_name} on table: {table_name} must be smaller than max value. Min is {min_value}, and max is {max_value}".format(
                    column_name=column_name,
                    table_name=table_name,
                    min_value=min_value,
                    max_value=max_value,
                )
            )
            return None, []

        bin_width = (max_value - min_value) / number_of_intervals

        if bin_width.is_integer() and column_type == "integer":
            bin_width = int(bin_width)
            min_value = int(min_value)
            max_value = int(max_value)

        bins_list = [round(min_value + i * bin_width, 2) for i in range(0, number_of_bins)]

        field_clauses = []
        for i in range(0, number_of_bins):
            lower_bound = "" if i == 0 else f"{bins_list[i]} <= value_"
            upper_bound = "" if i == number_of_bins - 1 else f"value_ < {bins_list[i + 1]}"
            optional_and = "" if lower_bound == "" or upper_bound == "" else " AND "
            field_clauses.append(f"SUM(CASE WHEN {lower_bound}{optional_and}{upper_bound} THEN frequency_ END)")

        fields = ",\n ".join(field_clauses)

        value_frequencies_cte = self.profiling_sql_value_frequencies_cte(table_name, column_name)

        sql = dedent(
            f"""
            WITH
                {value_frequencies_cte}
            SELECT {fields}
            FROM value_frequencies"""
        )
        return sql, bins_list

    def sql_test_connection(self) -> str:
        return "SELECT 1"

    ######################
    # Query Execution
    ######################

    def get_row_counts_all_tables(
        self,
        include_tables: list[str] | None = None,
        exclude_tables: list[str] | None = None,
        query_name: str | None = None,
    ) -> dict[str, int]:
        """
        Returns a dict that maps table names to row counts.
        """
        sql = self.sql_get_table_names_with_count(include_tables=include_tables, exclude_tables=exclude_tables)
        if sql:
            query = Query(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=query_name or "get_row_counts_all_tables",
                sql=sql,
            )
            query.execute()
            return {self._optionally_quote_table_name_from_meta_data(row[0]): row[1] for row in query.rows}

        # Single query to get the metadata not available, get the counts one by one.
        all_tables = self.get_table_names(include_tables=include_tables, exclude_tables=exclude_tables)
        result = {}
        for table_name in all_tables:
            table_count = self.get_table_row_count(table_name)
            if table_count:
                result[table_name] = table_count

        return result

    def get_table_row_count(self, table_name: str) -> int | None:
        """Deprecated, use get_row_counts_all_tables whenever possible."""
        query_name_str = f"get_row_count_{table_name}"
        query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name=query_name_str,
            sql=self.sql_get_table_count(self.quote_table(table_name)),
        )
        query.execute()
        if query.rows:
            return query.rows[0][0]
        return None

    def get_table_names(
        self,
        filter: str | None = None,
        include_tables: list[str] = [],
        exclude_tables: list[str] = [],
        query_name: str | None = None,
    ) -> list[str]:
        if not include_tables and not exclude_tables:
            return []
        sql = self.sql_find_table_names(filter, include_tables, exclude_tables)
        query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name=query_name or "get_table_names",
            sql=sql,
        )
        query.execute()
        table_names = [self._optionally_quote_table_name_from_meta_data(row[0]) for row in query.rows]
        return table_names

    def _optionally_quote_table_name_from_meta_data(self, table_name: str) -> str:
        """
        To be used by all table names coming from metadata queries.  Quotes are added if needed if the table
        doesn't match the default casify rules or if whitespaces are present.
        The table_name is returned unquoted otherwise.
        """
        # if the table name needs quoting
        if table_name != self.default_casify_table_name(table_name):
            return self.quote_table(table_name)
        elif self.quote_tables:
            return self.quote_table(table_name)
        else:
            # return the bare table name
            return table_name

    def analyze_table(self, table: str):
        if self.sql_analyze_table(table):
            QueryWithoutResults(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=f"analyze_{table}",
                sql=self.sql_analyze_table(table),
            ).execute()

    def _create_table_prefix(self):
        """
        Use
            * self.schema
            * self.database
            * self.quote_table(unquoted_table_name)
        to compose the table prefix to be used in Soda Core queries.  The returned table prefix
        should not include the dot (.) and can optionally be None.  Consider quoting as well.
        Examples:
            return None
            return self.schema
            return self.database
            return f'"{self.database}"."{self.schema}"'
        """
        return self.schema

    def update_schema(self, schema_name):
        self.schema = schema_name
        self.table_prefix = self._create_table_prefix()

    def qualified_table_name(self, table_name: str) -> str:
        """
        table_name can be quoted or unquoted
        """
        if self.table_prefix:
            return f"{self.table_prefix}.{table_name}"
        return table_name

    def quote_table_declaration(self, table_name: str) -> str:
        return self.quote_table(table_name=table_name)

    def is_quoted(self, table_name: str) -> bool:
        return (table_name.startswith('"') and table_name.endswith('"')) or (
            table_name.startswith("'") and table_name.endswith("'")
        )

    def quote_table(self, table_name: str) -> str:
        return f'"{table_name}"'

    def quote_column_declaration(self, column_name: str) -> str:
        return self.quote_column(column_name)

    def quote_column(self, column_name: str) -> str:
        return f'"{column_name}"'

    def get_sql_type_for_create_table(self, data_type: str) -> str:
        if data_type in self.SQL_TYPE_FOR_CREATE_TABLE_MAP:
            return self.SQL_TYPE_FOR_CREATE_TABLE_MAP.get(data_type)
        else:
            return data_type

    def get_sql_type_for_schema_check(self, data_type: str) -> str:
        data_source_type = self.SQL_TYPE_FOR_SCHEMA_CHECK_MAP.get(data_type)
        if data_source_type is None:
            raise NotImplementedError(
                f"Data type {data_type} is not mapped in {type(self)}.SQL_TYPE_FOR_SCHEMA_CHECK_MAP"
            )
        return data_source_type

    def literal(self, o: object):
        if o is None:
            return "NULL"
        elif isinstance(o, Number):
            return self.literal_number(o)
        elif isinstance(o, str):
            return self.literal_string(o)
        elif isinstance(o, datetime):
            return self.literal_datetime(o)
        elif isinstance(o, date):
            return self.literal_date(o)
        elif isinstance(o, list) or isinstance(o, set) or isinstance(o, tuple):
            return self.literal_list(o)
        elif isinstance(o, bool):
            return self.literal_boolean(o)
        raise RuntimeError(f"Cannot convert type {type(o)} to a SQL literal: {o}")

    def literal_number(self, value: Number):
        if value is None:
            return None
        return str(value)

    def literal_string(self, value: str):
        if value is None:
            return None
        return "'" + self.escape_string(value) + "'"

    def literal_list(self, l: list):
        if l is None:
            return None
        return "(" + (",".join([self.literal(e) for e in l])) + ")"

    def literal_date(self, date: date):
        date_string = date.strftime("%Y-%m-%d")
        return f"DATE '{date_string}'"

    def literal_datetime(self, datetime: datetime):
        return f"'{datetime.isoformat()}'"

    def literal_boolean(self, boolean: bool):
        return "TRUE" if boolean is True else "FALSE"

    def expr_count_all(self) -> str:
        return self.expr_count("*")

    def expr_count_conditional(self, condition: str):
        return self.expr_count(self.expr_conditional(condition, "1"))

    def expr_conditional(self, condition: str, expr: str):
        return f"CASE WHEN {condition} THEN {expr} END"

    def expr_count(self, expr):
        return f"COUNT({expr})"

    def expr_distinct(self, expr):
        return f"DISTINCT({expr})"

    def expr_length(self, expr):
        return f"LENGTH({expr})"

    def expr_min(self, expr):
        return f"MIN({expr})"

    def expr_max(self, expr):
        return f"MAX({expr})"

    def expr_avg(self, expr):
        return f"AVG({expr})"

    def expr_sum(self, expr):
        return f"SUM({expr})"

    def expr_regexp_like(self, expr: str, regex_pattern: str):
        return f"REGEXP_LIKE({expr}, '{regex_pattern}')"

    def expr_regex_condition(self, expr: str, condition: str) -> str:
        return condition.format(expr=expr)

    def get_default_format_expression(self, expr: str, format: str, location: Location | None = None) -> str | None:
        default_expression = self.DEFAULT_FORMAT_EXPRESSIONS.get(format)
        if default_expression:
            return self.expr_regex_condition(expr, default_expression)

        default_format = self.DEFAULT_FORMATS.get(format)
        if default_format:
            return self.expr_regexp_like(expr, self.escape_regex(default_format))

        # TODO move this to a validate step between configuration parsing and execution so that it can be validated without running
        self.logs.error(
            f"Format {format} is not supported.",
            location=location,
        )

        return None

    def expr_in(self, left: str, right: str):
        return f"{left} IN {right}"

    def cast_text_to_number(self, column_name, validity_format: str):
        """Cast string to number

        - first regex replace removes extra chars, keeps: "digits + - . ,"
        - second regex changes "," to "."
        - Nullif makes sure that if regexes return empty string then Null is returned instead
        """
        regex = self.escape_regex(r"'[^-0-9\.\,]'")
        return f"CAST(NULLIF(REGEXP_REPLACE(REGEXP_REPLACE({column_name}, {regex}, ''{self.regex_replace_flags()}), ',', '.'{self.regex_replace_flags()}), '') AS {self.SQL_TYPE_FOR_CREATE_TABLE_MAP[DataType.DECIMAL]})"

    def regex_replace_flags(self) -> str:
        return ", 'g'"

    def escape_string(self, value: str):
        return re.sub(r"(\\.)", r"\\\1", value)

    def escape_regex(self, value: str):
        return value

    def get_max_aggregation_fields(self):
        """
        Max number of fields to be aggregated in 1 aggregation query
        """
        return 50

    def connect(self):
        """
        Subclasses use self.data_source_properties to initialize self.connection with a PEP 249 connection

        Any BaseException may be raised in case of errors.
        The caller of this method will catch the exception and add an error log to the scan.
        """
        raise NotImplementedError(f"TODO: Implement {type(self)}.connect()")

    def fetchall(self, sql: str):
        # TODO: Deprecated - not used, use Query object instead.
        try:
            cursor = self.connection.cursor()
            try:
                self.logs.info(f"Query: \n{sql}")
                cursor.execute(sql)
                return cursor.fetchall()
            finally:
                cursor.close()
        except BaseException as e:
            self.logs.error(f"Query error: {e}\n{sql}", exception=e)
            self.query_failed(e)

    def is_connected(self):
        return self.connection is not None

    def disconnect(self) -> None:
        if self.connection:
            self.connection.close()
            # self.connection = None is used in self.is_connected
            self.connection = None

    def commit(self):
        self.connection.commit()

    def query_failed(self, e: BaseException):
        self.rollback()

    def rollback(self):
        self.connection.rollback()

    def default_casify_sql_function(self) -> str:
        """Returns the sql function to use for default casify."""
        return "lower"

    def default_casify_system_name(self, identifier: str) -> str:
        """Formats database/schema/etc identifier to e.g. a default case for a given data source."""
        return identifier.lower()

    def default_casify_table_name(self, identifier: str) -> str:
        """Formats table identifier to e.g. a default case for a given data source."""
        return identifier

    def default_casify_column_name(self, identifier: str) -> str:
        """Formats column identifier to e.g. a default case for a given data source."""
        return identifier

    def default_casify_type_name(self, identifier: str) -> str:
        """Formats type identifier to e.g. a default case for a given data source."""
        return identifier

    def safe_connection_data(self):
        """Return non-critically sensitive connection details.

        Useful for debugging and telemetry.
        """
        # to be overridden by subclass

    def generate_hash_safe(self):
        """Generates a safe hash from non-sensitive connection details.

        Useful for debugging, identifying data sources anonymously and tracing.
        """
        data = self.safe_connection_data()

        return self.hash_data(data)

    def hash_data(self, data) -> str:
        """Hash provided data using a non-reversible hashing algorithm."""
        encoded = json.dumps(data, sort_keys=True).encode()
        return hashlib.sha256(encoded).hexdigest()

    def test(self, sql):
        import logging
        import textwrap

        from soda.sampler.log_sampler import LogSampler
        from soda.sampler.sample_schema import SampleColumn

        cursor = self.connection.cursor()
        try:
            indented_sql = textwrap.indent(text=sql, prefix="  #   ")
            logging.debug(f"  # Query: \n{indented_sql}")
            cursor.execute(sql)
            rows = cursor.fetchall()

            columns = SampleColumn.create_sample_columns(cursor.description, self)
            table, row_count, col_count = LogSampler.pretty_print(rows, columns)
            logging.debug(f"  # Query result: \n{table}")

        except Exception as e:
            logging.error(f"Error: {e}", e)

        finally:
            cursor.close()

    def sql_groupby_count_categorical_column(
        self,
        select_query: str,
        column_name: str,
        limit: int | None = None,
    ) -> str:
        cte = select_query.replace("\n", " ")
        # delete multiple spaces
        cte = re.sub(" +", " ", cte)
        sql = dedent(
            f"""
                WITH processed_table AS (
                    {cte}
                )
                SELECT
                    {column_name}
                    , {self.expr_count_all()} AS frequency
                FROM processed_table
                GROUP BY {column_name}
            """
        )
        sql += f"LIMIT {limit}" if limit else ""
        return dedent(sql)

    def sql_select_column_with_filter_and_limit(
        self,
        column_name: str,
        table_name: str,
        filter_clause: str | None = None,
        sample_clause: str | None = None,
        limit: int | None = None,
    ) -> str:
        """
        Returns a SQL query that selects a column from a table with optional filter, sample query
        and limit.
        """
        filter_clauses_str = f"\n WHERE {filter_clause}" if filter_clause else ""
        sample_clauses_str = f"\n {sample_clause}" if sample_clause else ""
        limit_str = f"\n LIMIT {limit}" if limit else ""
        sql = f"SELECT \n" f"  {column_name} \n" f"FROM {table_name}{sample_clauses_str}{filter_clauses_str}{limit_str}"
        return sql

    def expr_false_condition(self):
        return "FALSE"



from __future__ import annotations

from typing import Any, ClassVar, List, Optional, Union

from pydantic import BaseModel, ConfigDict, ValidationError, field_validator
from soda.common.logs import Logs
from soda.sodacl.change_over_time_cfg import ChangeOverTimeCfg
from soda.sodacl.location import Location
from soda.sodacl.metric_check_cfg import MetricCheckCfg
from soda.sodacl.missing_and_valid_cfg import MissingAndValidCfg
from soda.sodacl.threshold_cfg import ThresholdCfg


class ADBaseModel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")
    logger: ClassVar[Logs]
    location: ClassVar[Location]

    @classmethod
    def create_instance(cls, logger: Logs, location: Location, **kwargs: Any) -> ADBaseModel | None:
        try:
            return cls(**kwargs)
        except ValidationError as e:
            for error in e.errors():
                # only keep string instance field names
                field_names = [loc for loc in error["loc"] if isinstance(loc, str)]
                field_name = field_names[-1]
                field_value = error.get("input")  # Get the provided value
                if error.get("type") == "missing":
                    logger.error(
                        f"Anomaly Detection Parsing Error: Missing field '{field_name}' at {location}."
                        f" Configure the required field in your SodaCL file."
                    )
                elif error.get("type") == "extra_forbidden":
                    logger.error(
                        f"Anomaly Detection Parsing Error: Extra field '{field_name}' at {location}."
                        f" Remove the field from your SodaCL file."
                    )
                elif error.get("type") == "value_error":
                    logger.error(
                        f"Anomaly Detection Parsing Error: Not allowed value for field "
                        f"'{field_name}' at {location}. "
                        f"{error['msg']}."
                    )
                else:
                    logger.error(
                        "Anomaly Detection Parsing Error: Unexpected value "
                        f"'{field_value}' for field '{field_name}' at {location}. "
                        f"{error['msg']}."
                    )
            return None
        except ValueError as e:
            logger.error(f"Error while parsing {cls.__name__} at {location}:\n{e}")
            return None


class ProphetDefaultHyperparameters(ADBaseModel):
    growth: str = "linear"
    changepoints: Any = None
    n_changepoints: int = 25
    changepoint_range: float = 0.8
    yearly_seasonality: Any = "auto"
    weekly_seasonality: Any = "auto"
    daily_seasonality: Any = "auto"
    holidays: Any = None
    seasonality_mode: str = "multiplicative"  # Tuned
    seasonality_prior_scale: float = 0.01  # Tuned
    holidays_prior_scale: float = 10.0
    changepoint_prior_scale: float = 0.001  # Tuned
    mcmc_samples: int = 0
    interval_width: float = 0.999  # Tuned
    uncertainty_samples: int = 1000
    stan_backend: Any = None
    scaling: str = "absmax"
    holidays_mode: Any = None


class ProphetMAPEProfileHyperparameters(ProphetDefaultHyperparameters):
    seasonality_prior_scale: float = 0.1  # Tuned
    changepoint_prior_scale: float = 0.1  # Tuned


class ProphetParameterGrid(ADBaseModel):
    growth: List[str] = ["linear"]
    changepoints: List[Any] = [None]
    n_changepoints: List[int] = [25]
    changepoint_range: List[float] = [0.8]
    yearly_seasonality: List[Any] = ["auto"]
    weekly_seasonality: List[Any] = ["auto"]
    daily_seasonality: List[Any] = ["auto"]
    holidays: List[Any] = [None]
    seasonality_mode: List[str] = ["multiplicative"]  # Non default
    seasonality_prior_scale: List[float] = [0.01, 0.1, 1.0, 10.0]  # Non default
    holidays_prior_scale: List[float] = [10.0]
    changepoint_prior_scale: List[float] = [0.001, 0.01, 0.1, 0.5]  # Non default
    mcmc_samples: List[int] = [0]
    interval_width: List[float] = [0.999]  # Non default
    stan_backend: List[Any] = [None]
    scaling: List[str] = ["absmax"]
    holidays_mode: List[Any] = [None]


class ProphetDynamicHyperparameters(ADBaseModel):
    objective_metric: Union[str, List[str]]
    parallelize_cross_validation: bool = True
    cross_validation_folds: int = 5
    frequency: int = 10
    parameter_grid: ProphetParameterGrid = ProphetParameterGrid()

    @field_validator("objective_metric", mode="before")
    @classmethod
    def metric_is_allowed(cls, v: str | List[str]) -> str | List[str]:
        allowed_metrics = ["mse", "rmse", "mae", "mape", "mdape", "smape", "coverage"]
        error_message = (
            "objective_metric: '{objective_metric}' is not allowed. "
            "Please choose from 'mse', 'rmse', 'mae', 'mape', 'mdape', 'smape', 'coverage'."
        )
        if isinstance(v, List):
            v = [metric.lower() for metric in v]
            for metric in v:
                if metric not in allowed_metrics:
                    raise ValueError(error_message.format(objective_metric=metric))
        else:
            if v.lower() not in ["mse", "rmse", "mae", "mape", "mdape", "smape", "coverage"]:
                raise ValueError(error_message.format(objective_metric=v))
        return v


class ProphetCustomHyperparameters(ADBaseModel):
    custom_hyperparameters: ProphetDefaultHyperparameters = ProphetDefaultHyperparameters()


class ProphetHyperparameterProfiles(ADBaseModel):
    profile: ProphetCustomHyperparameters = ProphetCustomHyperparameters()

    @field_validator("profile", mode="before")
    def set_profile(cls, v: Union[str, ProphetCustomHyperparameters]) -> ProphetCustomHyperparameters:
        if isinstance(v, str):
            v = v.lower()
            if v == "mape":
                return ProphetCustomHyperparameters(custom_hyperparameters=ProphetMAPEProfileHyperparameters())
            elif v == "coverage":
                return ProphetCustomHyperparameters()
            else:
                raise ValueError(f"Profile: '{v}' is not allowed. " f"Please choose from 'MAPE' or 'coverage'.")
        else:
            return v


class HyperparameterConfigs(ADBaseModel):
    static: ProphetHyperparameterProfiles = ProphetHyperparameterProfiles()
    dynamic: Optional[ProphetDynamicHyperparameters] = None


class ModelConfigs(ADBaseModel):
    type: str = "prophet"
    holidays_country_code: Optional[str] = None
    hyperparameters: HyperparameterConfigs = HyperparameterConfigs()


class TrainingDatasetParameters(ADBaseModel):
    frequency: str = "auto"
    aggregation_function: str = "last"
    window_length: int = 1000


class SeverityLevelParameters(ADBaseModel):
    warning_ratio: float = 0.1
    min_confidence_interval_ratio: float = 0.001

    @field_validator("warning_ratio", "min_confidence_interval_ratio")
    def check_ratio(cls, v: float) -> float:
        if not 0 <= v <= 1:
            raise ValueError(f"Value must be between 0 and 1, but got {v}")
        return v


class AnomalyDetectionMetricCheckCfg(MetricCheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: str | None,
        location: Location,
        name: str | None,
        metric_name: str,
        metric_args: List[object] | None,
        missing_and_valid_cfg: MissingAndValidCfg | None,
        filter: str | None,
        condition: str | None,
        metric_expression: str | None,
        metric_query: str | None,
        change_over_time_cfg: ChangeOverTimeCfg | None,
        fail_threshold_cfg: ThresholdCfg | None,
        warn_threshold_cfg: ThresholdCfg | None,
        model_cfg: ModelConfigs,
        training_dataset_params: TrainingDatasetParameters,
        severity_level_params: SeverityLevelParameters,
        take_over_existing_anomaly_score_check: bool = False,
        is_automated_monitoring: bool = False,
        samples_limit: int | None = None,
        samples_columns: List | None = None,
    ):
        super().__init__(
            source_header,
            source_line,
            source_configurations,
            location,
            name,
            metric_name,
            metric_args,
            missing_and_valid_cfg,
            filter,
            condition,
            metric_expression,
            metric_query,
            change_over_time_cfg,
            fail_threshold_cfg,
            warn_threshold_cfg,
            samples_limit=samples_limit,
        )
        self.is_automated_monitoring = is_automated_monitoring
        self.model_cfg = model_cfg
        self.training_dataset_params = training_dataset_params
        self.severity_level_params = severity_level_params
        self.take_over_existing_anomaly_score_check = take_over_existing_anomaly_score_check



class SchemaComparator:
    def __init__(self, historic_schema, measured_schema):
        self.historic_schema = historic_schema
        self.measured_schema = measured_schema

        self.schema_column_deletions = []
        self.schema_column_additions = []
        self.schema_column_type_changes = {}
        self.schema_column_index_changes = {}

        self.__compute_schema_changes()

    def __compute_schema_changes(self):
        column_types = {column["name"]: column["type"] for column in self.measured_schema}
        column_indexes = {column["name"]: index for index, column in enumerate(self.measured_schema)}

        previous_column_types = {column["name"]: column["type"] for column in self.historic_schema}
        previous_column_indexes = {column["name"]: index for index, column in enumerate(self.historic_schema)}
        for previous_column in previous_column_types:
            if previous_column not in column_types:
                self.schema_column_deletions.append(previous_column)
        for column in column_types:
            if column not in previous_column_types:
                self.schema_column_additions.append(column)
        for column in column_types:
            if column in previous_column_types:
                type = column_types[column]
                previous_type = previous_column_types[column]
                if type != previous_type:
                    self.schema_column_type_changes[column] = {
                        "previous_type": previous_type,
                        "new_type": type,
                    }
                index = column_indexes[column]
                previous_index = previous_column_indexes[column]
                if index != previous_index:
                    self.schema_column_index_changes[column] = {
                        "previous_index": previous_index,
                        "new_index": index,
                    }



class Lazy:
    def __init__(self):
        self.object = None

    def is_set(self) -> bool:
        return self.object is not None

    def get(self):
        return self.object

    def set(self, o):
        self.object = o
